/*****************************************************************************
 * nnet.hpp - Header file for basic neural network library (C++)             *
 * (libnnet.so) or (libnnet.a) - Version 0.1                                 *
 * ------------------------------------------------------------------------- *
 * This library contains the basic classes and methods for the               *
 * implementation of simple neural networks based on the mathematical neuron *
 * model proposed by McCulloch and Pitts.                                    *
 * ------------------------------------------------------------------------- *
 * Copyright 2003 by Rosario Piro (mailto: devfaust@users.sourceforge.net)   *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

#ifndef _NNET_H_
#define _NNET_H_

//*********************
//*** include files ***
//*********************
#include <extstdlib.h>
#include <stdclassdef.hpp>
#include <vector>
#include <map>
#include <values.h>

//***************************
//*** preprocessor macros ***
//***************************
#define NNETLIB "devfaust Neural Network Library"
#define NNETVER "Version 0.1"

#define DEF_INIT_WEIGHT 0.2 * rand()/RAND_MAX - 0.1

//**************************
//*** class declarations ***
//**************************

namespace nnet {

  // class declarations and typedefs:

  class Synapse;
  class Neuron;
  class InputTerminal;  // derived from class Neuron

  class NNet;     // generic neural network
  class FFNNet;   // feed-forward network, derived from class NNet
  class FFNNetBP; // feed-forward network with back-propagation
                  // derived from FFNNet

  typedef map<unsigned int,Synapse*>        synapse_map_t;
  typedef map<unsigned int,Neuron*>         neuron_map_t;
  typedef map<unsigned int,InputTerminal*>  interm_map_t;

  typedef vector<Neuron*>  n_vect_t;      // neuron layer for feed-forward nets
  typedef vector<Synapse*> syn_vect_t;    // vector of synapses

  typedef vector<double> train_ex_in_t;   // input vector of a training example
  typedef vector<double> train_ex_out_t;  // outp. vector of a training example
  typedef pair<train_ex_in_t*,train_ex_out_t*> train_ex_t; // training example
  typedef vector<train_ex_t> train_set_t; // set of training examples

  // class definitions:

  // *********************
  // *** class Synapse ***
  // *********************

  // The Synapse class represents a weighted unidirectional connection
  // between two neurons.
  // A neural network is trained by modifying the weights of the connection
  // between its neurons.
  class Synapse { 

    // private and protected data members with accessor methods
    // (type, privateName, PublicName) -> GetPublicName(), SetPublicName(type):

    PROT_ATTRIBUTE_GET(unsigned int, _synID, SynID) // id number of synapse

    // weight of the connection, SetWeight() is defined later.
    PRIV_ATTRIBUTE_GET(double, _weight, Weight)

    // presynaptic (incoming) neuron 
    PRIV_ATTRIBUTE_GET_SET(Neuron *, _from, From) 

    // postsynaptic (outgoing) neuron 
    PRIV_ATTRIBUTE_GET_SET(Neuron *, _to, To)

  public:
    // Constructors
    Synapse(unsigned int sID, double wgt, Neuron *nf = NULL,
	    Neuron *nt = NULL) :
      _synID(sID), _weight(wgt), _from(nf), _to(nt) {}

    Synapse(unsigned int sID, Neuron *nf = NULL, Neuron *nt = NULL) :
       _synID(sID), _from(nf), _to(nt) {
      _weight = DEF_INIT_WEIGHT;}

    // additional accessor method (non standard):
    void SetWeight(double wgt);  // GetWeight() is defined by macro.

    // method for synaptical function: multiplies weight with input
    double WeightedVal();

  }; // class Synapse ...


  // ********************
  // *** class Neuron ***
  // ********************

  // The Neuron class implements a nercous cellar unit that sums the pulses
  // received from the presynaptic (incoming) neurons connected to its
  // dentrites (linear function) and transforms them (non-linear function)
  // according to a fixed (non-programmable) activation function.
  // If the resulting potential reaches a threshold the neuron transmits a
  // pulse (action potential) via it's axon to it's postsynaptic (outgoing)
  // neurons. 
  class Neuron {

    // protected data members with accessor methods:
    // (type, privateName, PublicName) -> GetPublicName(), SetPublicName(type):

    PROT_ATTRIBUTE_GET(unsigned int, _neuronID, NeuronID) // id of neuron

    // sum of all input values from presynaptic neurons.
    PROT_ATTRIBUTE_GET(double, _inSum, InSum)

    // action potential (output) to postsnaptic neurons.
    PROT_ATTRIBUTE_GET(double, _outVal, OutputVal)

    // TRUE if input values have changed (new sum has to be calculated).
    PROT_ATTRIBUTE_GET_SET(bool, _inChange, InputChange) 

    // partial derivate for the back-propagation training algorithm.
    PRIV_ATTRIBUTE_GET(double, _delta, DeltaVal)

    // vector of synapses from presynaptic (incoming) neurons.
    PROT_ATTRIBUTE_GET_REF(syn_vect_t, _inSynVect, InSynVect)

    // vector of synapses from postsynaptic (outgoing) neurons.
    PROT_ATTRIBUTE_GET_REF(syn_vect_t, _outSynVect, OutSynVect)

  public:
    // Constructors
    Neuron(unsigned int nID) :
      _neuronID(nID), _inSum(0), _outVal(0), _inChange(TRUE), _delta(0),
      _inSynVect(), _outSynVect() {}                  // default

    Neuron(unsigned int nID, syn_vect_t outSV) : 
      _neuronID(nID), _inSum(0), _outVal(0), _inChange(TRUE), _delta(0),
      _inSynVect(), _outSynVect(outSV) {}             // for input terminals

    Neuron(unsigned int nID, syn_vect_t inSV, syn_vect_t outSV) :
      _neuronID(nID), _inSum(0), _outVal(0), _inChange(TRUE), _delta(0),
      _inSynVect(inSV),_outSynVect(outSV) {}        // for internal neutrons

    Neuron(unsigned int nID, double sum, double out, bool inch, double delta) :
      _neuronID(nID), _inSum(sum), _outVal(out), _inChange(inch),
      _delta(delta), _inSynVect(), _outSynVect() {}

    // Destructor
    virtual ~Neuron() {}

    // sums all input values and calculates output
    virtual double AxonOutput();

    // activation function (computes on _inSum)
    // This function is virtual. The library however implements the most
    // common activation function for neural network experiments: the sigmoid
    // function. 
    virtual double ActFunction();

    // derived activation function for back propagation algorithms (on _inSum)
    virtual double DerivActFunction();

    // returns the partial derivate for BP (on _inSum), error is passed only
    // for output terminals
    virtual double Delta(double error = MAXDOUBLE);

    // methods for neuron connections (synaptical links):
    void AddInSyn(Synapse *newInSyn);      // adds a new input synapse
    bool IsInSyn(const Synapse *syn);      // returns true if syn is input
    bool IsInSyn(unsigned int synID);      // returns true is synID is input
    Synapse *GetInSyn(unsigned int synID); // get ptr. to input syn with synID
    bool DelInSyn(Synapse *oldsyn);        // removes input synapse oldsyn
    bool DelInSyn(unsigned int synID);     // removes input synapse with synID

    void AddOutSyn(Synapse *newOutSyn);    // adds a new output synapse
    bool IsOutSyn(const Synapse *syn);     // returns true if syn is output
    bool IsOutSyn(unsigned int synID);     // returns true is synID is output
    Synapse *GetOutSyn(unsigned int synID);// get ptr. with outp.syn with synID
    bool DelOutSyn(Synapse *oldsyn);       // removes output synapse oldsyn
    bool DelOutSyn(unsigned int synID);    // removes output synapse with synID

    void AddInNeuron(Neuron &newIn, unsigned int synID,
		     double weight = MAXDOUBLE); // adds as input, creating
                                                 // the synapse, in case of
                                                 // MAXDOUBLE, weight will be
                                                 // set to DEF_INIT_WEIGHT
    void AddOutNeuron(Neuron &newOut, unsigned int synID,
		      double weight = MAXDOUBLE); // adds as output, creating
                                                  // the synapse, in case of
                                                  // MAXDOUBLE, weight will be
                                                  // set to DEF_INIT_WEIGHT
  }; // class Neuron ...


  // *******************************************
  // *** class InputTerminal : public Neuron ***
  // *******************************************

  // As input terminals to the neural network we use derived Neurons
  // This requires an additional data member and different implementations
  // for some of the virtual functions.
  // Input terminal do not compute an activation function but pass the input
  // value unaltered to the axon.
  class InputTerminal : public Neuron {
  protected:
    double _input; // input value for user interface to the neural network

  public:
    // Constructors
    InputTerminal(unsigned int itID, double in = 0): Neuron(itID),_input(in) {}
    InputTerminal(unsigned int itID, syn_vect_t outSV, double in = 0) :
      Neuron(itID, outSV), _input(in) {}
    // Destructor
    virtual ~InputTerminal() {}

    // Accessor (non standard)
    void SetInput(double in);

    // propriate definitons of virtual functions
    double AxonOutput();

  }; // class InputTerminal ...


  // ******************
  // *** class NNet ***
  // ******************

  // This class represents a generic neural network. It is intended as base
  // class for more specific network architectures.
  class NNet {

    // protected data members and their accessor methods:
    // (type, privateName, PublicName) -> GetPublicName(), SetPublicName(type):

    PROT_ATTRIBUTE_GET(unsigned int, _NNetID, NNetID) // ID of neural netw.

    // Input terminals: maps neuron IDs to pointers to the input terminals.
    PROT_ATTRIBUTE_GET_REF(interm_map_t, _ITMap, ITMap)

    // Computing neurons: maps neuron IDs to pointers to the computing neurons.
    PROT_ATTRIBUTE_GET_REF(neuron_map_t, _NMap, NMap)

    // Output terminals: maps neuron IDs to pointers to the input terminals.
    // (Important: output terminals are also computing neurons and thus
    //  part of _NMap)
    PROT_ATTRIBUTE_GET_REF(neuron_map_t, _OutMap, OutMap)

    // Synapses: maps synapse IDs to pointers to the synapses.
    PROT_ATTRIBUTE_GET_REF(synapse_map_t, _SynMap, SynMap)

    // for automatic increment of neuronIDs, contains current number of
    // neurons (input terminals and computing neurons)
    PROT_ATTRIBUTE_GET(unsigned int, _currNID, CurrentNeuronID)

    // for automatic increment of synIDs, contains current number of synapses
    PROT_ATTRIBUTE_GET(unsigned int, _currSID, CurrentSynID)

  public:
    // Constructors:
    NNet(unsigned int nID) : 
      _NNetID(nID), _ITMap(), _NMap(), _OutMap(), _SynMap(),
      _currNID(0), _currSID(0) {}

    NNet(unsigned int nID, interm_map_t itmap, neuron_map_t nmap,
	 neuron_map_t outmap, synapse_map_t synmap);

    // Destructor:
    virtual ~NNet();

    // This function inserts an InputTerminal to the network and returns
    // true if successful.
    // The function is virtual because it might be necessary to redefine
    // it for specific network structures.
    // If the neuron ID (nID) is not specified, the neuron ID will depend
    // on _currNID (default). 
    virtual bool InsertInTerm(InputTerminal *inT, unsigned int nID = UINT_MAX);

    // CreateInTerm() creates a new InputTerminal for the network and
    // returns true if successful.
    // If the neuron ID (nID) is not specified, the neuron ID will depend
    // on _currNID (default). 
    bool CreateInTerm(unsigned int nID = UINT_MAX, double input = 0);

    // This function inserts a computing neuron to the network and returns
    // true if successful.
    // The function is virtual because it might be necessary to redefine
    // it for specific network structures.
    // If the neuron ID (nID) is not specified, the neuron ID will depend
    // on _currNID (default). 
    virtual bool InsertNeuron(Neuron *N, unsigned int nID = UINT_MAX);

    // CreateNeuron() creates a new computing neuron for the network and
    // returns true if successful.
    // If the neuron ID (nID) is not specified, the neuron ID will depend
    // on _currNID (default). 
    bool CreateNeuron(unsigned int nID = UINT_MAX, double inSum = 0,
		      double outVal = 0, bool inCh = TRUE, double delta = 0);

    // This function declares a computing neuron that is already part of the
    // network to be an output terminal and returns true if successful.
    // The function is virtual because it might be necessary to redefine
    // it for specific network structures.
    virtual bool MakeOutput(unsigned int nID);

    // CreateOutput() creates a new output terminal for the network and
    // returns true if successful. Important: Output terminals are
    // computing neurons as well!
    // If the neuron ID (nID) is not specified, the neuron ID will depend
    // on _currNID (default). 
    bool CreateOutput(unsigned int nID = UINT_MAX);

    // This function creates a synapse between two neurons:
    // The function is virtual because it might be necessary to redefine
    // it for specific network structures. If successful true is returned.
    // If the synapse ID (synID) is not specified, the synapse ID will depend
    // on _currSID (default). 
    virtual bool Connect(unsigned int fromNID, unsigned int toNID,
			 unsigned int synID = UINT_MAX,
			 double weight = MAXDOUBLE);
                         // if weight is MAXDOUBLE, we'll use DEF_INIT_WEIGHT 

    // NNetInput writes a vector of input values to the input terminal,
    // returns true if successful.
    bool NNetInput(vector<double> &input);

    // NNetOutput() reads the nnet output to a vector of output values.
    bool NNetOutput(vector<double> &output);

    // NNetIterate() implements the iteration to generate the nnet output.
    // This function is virtual because it might depend by the structure
    // of the network, thus it might be necessary to overload it.
    virtual void NNetIterate();

    // BuildNNet() returns true if the creation was successful.
    // This is a virtual function since it depends on the net structure.
    // The NNet class does not implement a network with a particular
    // structure, thus we just return false (nothing build). It might seem
    // wise to declare BuildNNet() to be purely virtual, but this would
    // imply that NNet couldn't be used directly and had to be derived.
    virtual bool BuildNNet() { return FALSE; }

  }; // class NNet ...


  // **********************************
  // *** class FFNNet : public NNet ***
  // **********************************

  // This class represents a simple feed-forward neural network, that has a 
  // layered network structure i.e. syanptical links are build only from one 
  // layer to the next.
  class FFNNet : public NNet {

    // private data memebers and their accessor methods:
    // (type, privateName, PublicName) -> GetPublicName(), SetPublicName(type):

    // allows access to the computing neurons using the layered structure.
    // neurons can still be accessed with NNet::GetNMap()
    PROT_ATTRIBUTE_GET_REF(vector<n_vect_t*>, _nLayers, NeuronLayers)

    // if BuildNNet() is invoked without an argument, this vector is used for
    // initializing the net.
    PRIV_ATTRIBUTE_GET_SET_REF(vector<unsigned int>, _lWVec, LayerWidthVect)
  
  private:
    bool _netBuilt; // true if the layered structure has already been built.

  public:
    // Constructors:
    FFNNet(unsigned int nID) : NNet(nID), _nLayers(), _netBuilt(FALSE) {}

    FFNNet(unsigned int nID, vector<unsigned int> lWV) :
      NNet(nID), _nLayers(), _lWVec(lWV), _netBuilt(FALSE) {}

    // Destructor:
    virtual ~FFNNet();

    // redefinitions of virtual functions of class NNet:

    // BuildNNet() returns true if the creation was successful.
    bool BuildNNet();

    // this overloading BuildNNet instance takes the network structure as
    // an argument, copies it to _lWVec and then calls BuildNNet(void) instance
    bool BuildNNet(vector<unsigned int> &lWV);

    // NNetIterate() implements the iteration to generate the nnet output.
    void NNetIterate();

  }; // class FFNNet ...


  // **************************************
  // *** class FFNNetBP : public FFNNet ***
  // **************************************

  // This class is based on FFNNet, adding a back-propagation algorithm (Bryson
  // and Ho, 1969), that trains the neural network using a set of training
  // values (input/output pairs).
  class FFNNetBP : public FFNNet {

  public:
    // Constructor:
    FFNNetBP(unsigned int nID) : FFNNet(nID) {}

    FFNNetBP(unsigned int nID, vector<unsigned int> layWidthVec) :
      FFNNet(nID, layWidthVec) {}

    // Destructor:
    virtual ~FFNNetBP() {}

    // Back-Propagation algorithm: returns true if successful, otherwise
    // (number of maximum iterations was reached) it will return false.
    bool BackPropagate(const train_ex_t &tex, double learnrate,
		       double threshold, int maxIter);

    // TrainNNet() implements a wrapper to the BackPropagation algorithm using
    // a whole set of training examples instead of only one.
    // Terminates immediately and returns false if one of the BackPropagation()
    // calls returns false.
    // This is a virtual function since it depends on the net structure.
    virtual bool TrainNNet(train_set_t &tset, double learnrate,
			   double threshold, int maxIter);

  }; // class FFNNetBP ...


} // namespace nnet ...

#endif // #ifndef _NNET_H_
