/*****************************************************************************
 * nnet.cpp - basic neural network library (C++)                             *
 * (libnnet.so) or (libnnet.a) - Version 0.1                                 *
 * ------------------------------------------------------------------------- *
 * This library contains the basic classes and methods for the               *
 * implementation of simple neural networks.                                 *
 * ------------------------------------------------------------------------- *
 * Copyright 2003 by Rosario Piro (mailto: devfaust@users.sourceforge.net)   *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

#ifdef DEBUG
#define NNET_DEBUG
#endif

//*********************
//*** include files ***
//*********************
#include <nnet.hpp>
#include <cmath>
#ifdef NNET_DEBUG
  #include <iostream>
#endif

//***************************
//*** preprocessor macros ***
//***************************
#define _AF_ALPHA  2.0  // parameters for the default activation 
#define _AF_BETA   1.0  // function (sigmoid), that however is virtual
#define _AF_X0     0.0  // and thus can be redefined by classes derived from
#define _AF_Y0    -1.0  // class Neuron

//******************************
//*** class member functions ***
//******************************

namespace nnet {

  //**************************************
  //*** class Synapse member functions ***
  //**************************************


  // *** Synapse::SetWeight(double) ***
  // sets a new weight for synaptical connections and informs the receiving
  // neutron (_to) about the change.
  inline void Synapse::SetWeight(double wgt) {
    _weight = wgt;
    _to->SetInputChange(TRUE); // inform the receiving neuron about change
                   // the receiver needs to calculate the new summed input
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (Synapse::SetWeight = S:SW):" << endl 
	 << " (S:SW) SynID=" << _synID << ": synaptical weight set to "
	 << _weight << endl
	 << " (S:SW) setting input flag of neuron with NeuronID = " 
	 << _to->GetNeuronID() << endl;
    #endif
  } // Synapse::SetWeight(void)


  // *** Synapse::WeightedVal(void) ***
  // method for synaptical function: multiplies weight with input
  inline double Synapse::WeightedVal() {
    return _weight * (_from->AxonOutput());
  }


  //*************************************
  //*** class Neuron member functions ***
  //*************************************


  // *** Neuron::AxonOutput(void) *** (virtual)
  // AxonOutput sums all input values and determines the output value.
  // The function may be redefined for derived classes (necessary for example
  // for input terminals)
  double Neuron::AxonOutput() {

    if (_inChange) { // only if input values have been changed!

      _inSum=0; // sum all synaptical inputs
      for(syn_vect_t::iterator it=_inSynVect.begin(); it!=_inSynVect.end(); it++) {
	_inSum += (*it)->WeightedVal();
      }
      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (Neuron::AxonOutput = N:AO):" << endl
	   << " (N:AO) NeuronID="
	   << _neuronID << " new input sum: " << _inSum << endl;
      #endif

      _outVal = ActFunction(); // transform _inSum to output
      #ifdef NNET_DEBUG
      cerr << " (N:AO) -> new output value: " << _outVal << endl;
      #endif


      _inChange = FALSE; // make sure we recalculate only after changes occured

      // now inform the receiving neurons about the changed output
      // thus making sure that they will calculate their new output.
      for(syn_vect_t::iterator it = _outSynVect.begin();
	                      it != _outSynVect.end(); it++) {

	((*it)->GetTo())->SetInputChange(TRUE);
        #ifdef NNET_DEBUG
	cerr << " (N:AO) setting input flag of neuron with NeuronID = " 
	     << ((*it)->GetTo())->GetNeuronID() << endl;
        #endif

      } // for(vector ...

    } // if (_inChange) ...

    return _outVal;
  } // Neuron::AxonOutput(void) ...


  // *** Neuron::ActFunction(void) *** (virtual)
  // default activation function (sigmoid)...computes and returns neuron output
  // calculated on _inSum
  inline double Neuron::ActFunction() { 
    return _AF_ALPHA * (1.0/(1.0+exp(-_AF_BETA*(_inSum-_AF_X0))) ) + _AF_Y0;
  }


  // *** Neuron::DerivActFunction(void) *** (virtual)
  // derivate of the default activation function (doesn't require new sum)
  // needed for back propagation
  inline double Neuron::DerivActFunction() {
    double t = exp( -_AF_BETA*(_inSum - _AF_X0) );
    return _AF_ALPHA * _AF_BETA * t / ((1.0+t)*(1.0+t));
  }


  // *** Neuron::Delta(double=default) *** (virtual)
  // returns the partial derivate needed for the back-propagation algorithm.
  // This is a virtual function and ma be overloaded (overloaded
  // by output terminals!)
  double Neuron::Delta(double error) {
    // default value for error is MAXDOUBLE!

    _delta = 0;

    if(error == MAXDOUBLE) { // default! no error passed => no output terminal

      for(syn_vect_t::iterator it = _outSynVect.begin();
	  it != _outSynVect.end(); it++) { // sum weighted difference of the
                                           // computed and the expected output
	_delta += ((*it)->GetWeight()) * ((*it)->GetTo()->GetDeltaVal());
      } // for ...

      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (Neuron::Delta = N:D):" << endl
	   << " (N:D) delta of computing neuron (neuronID=" << _neuronID
	   << ")" << endl << " (N:D) weighted input sum = " << _delta;
      #endif

    } // if ...
    else { // output error passed to function, thus we have an output terminal

      _delta = error;

      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (Neuron::Delta = N:D):" << endl
	   << " (N:D) delta of output terminal (neuronID=" << _neuronID
	   << ")" << endl << " (N:D) output error = " << _delta;
      #endif

    }

    // multiply with result of derived activation function (on input _inSum):
    _delta *= DerivActFunction();

    #ifdef NNET_DEBUG
    cerr << " => delta = " << _delta << endl;
    #endif

    return _delta;
  } // Neuron::Delta(double=default)


  // *** Neuron::AddInSyn(Synapse *) ***
  // adds a new input synapse at the end of _inSynVect;
  inline void Neuron::AddInSyn(Synapse *newInSyn) {
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (Neuron::AddInSyn =N:AIS):" << endl
	 << " (N:AIS) adding new input synapse (synID="
	 << newInSyn->GetSynID() << ") to NeuronID=" << _neuronID
	 << flush << endl;
    #endif
    newInSyn->SetTo(this);            // pointer from synapse to neuron
    _inSynVect.push_back(newInSyn);   // pointer from neuron to synapse
  } // Neuron::AddInSyn(Synapse *)


  // *** Neuron::AddOutSyn(Synapse *) ***
  // adds a new output synapse at the end of _outSynVect;
  inline void Neuron::AddOutSyn(Synapse *newOutSyn) {
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (Neuron::AddOutSyn = N:AOS): " << endl
	 << " (N:AOS) adding new output synapse (synID="
	 << newOutSyn->GetSynID() << ") to NeuronID=" << _neuronID
	 << flush << endl;
    #endif
    _outSynVect.push_back(newOutSyn);  // pointer from neuron to synapse
    newOutSyn->SetFrom(this);          // pointer from synapse to neuron
  } // Neuron::AddOutSyn(Synapse *) ...


  // *** Neuron::IsInSyn(Synapse *) ***
  // returns true if the synapse syn is one of the input synapses,
  // otherwise it returns NULL.
  bool Neuron::IsInSyn(const Synapse *syn) {
    for (syn_vect_t::iterator it = _inSynVect.begin();
	 it != _inSynVect.end(); it++) {
      if( (*it)==syn ) // we found it.
	return TRUE;
    } // for ...
    return FALSE;
  } // Neuron::IsInSyn(Synapse *) ...


  // *** Neuron::IsInSyn(unsigned int) ***
  // returns true if the synapse with synID is one of the input synapses,
  // otherwise it returns NULL.
  inline bool Neuron::IsInSyn(unsigned int synID) {
    return IsInSyn(GetInSyn(synID)); // call other instance of IsInSyn()
  } // Neuron::IsInSyn(unsigned int) ...


  // *** Neuron::IsOutSyn(Synapse *) ***
  // returns true if the synapse syn is one of the output synapses,
  // otherwise returns NULL.
  bool Neuron::IsOutSyn(const Synapse *syn) {
    for (syn_vect_t::iterator it = _outSynVect.begin();
	 it != _outSynVect.end(); it++) {
      if( (*it)==syn )  // we found it.
	return TRUE;
    } // for ...
    return FALSE;
  } // Neuron::IsOutSyn(Synapse *) ...


  // *** Neuron::IsOutSyn(unsigned int) ***
  // returns true if the synapse with synID is one of the output synapses,
  // otherwise it returns NULL.
  inline bool Neuron::IsOutSyn(unsigned int synID) {
    return IsOutSyn(GetOutSyn(synID)); // call other instance of IsInSyn()
  } // Neuron::IsOutSyn(unsigned int) ...


  // *** Neuron::GetInSyn(unsigned int) ***
  // returns reference to the synapse with synID is one of the input synapses,
  // otherwise it returns NULL.
  Synapse *Neuron::GetInSyn(unsigned int synID) {
    for (syn_vect_t::iterator it = _inSynVect.begin();
	 it != _inSynVect.end(); it++) {
      if((*it)->GetSynID()==synID) // we found it.
	return *it;
    } // for ...
    return NULL;
  } // Neuron::GetInSyn(unsigned int) ...


  // *** Neuron::GetOutSyn(unsigned int) ***
  // returns reference to the synapse with synID is one of the output synapses,
  // otherwise returns NULL.
  Synapse *Neuron::GetOutSyn(unsigned int synID) {
    for (syn_vect_t::iterator it = _outSynVect.begin();
	 it != _outSynVect.end(); it++) {
      if((*it)->GetSynID()==synID)  // we found it.
	return *it;
    } // for ...
    return NULL;
  } // Neuron::GetOutSyn(unsigned int) ...


  // *** Neuron::DelInSyn(Synapse *) ***
  // searches for the synapse oldsyn in _inSynVect and removes it.
  // returns true if the synapse has been found and erased, false otherwise.
  bool Neuron::DelInSyn(Synapse *oldsyn) {

    // check if it's an input synapse:
    for(syn_vect_t::iterator it = _inSynVect.begin();
	                    it != _inSynVect.end(); it++){

      if((*it)==oldsyn) { // we found it.
        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (Neuron::DelInSyn = N:DIS): " << endl
	     << " (N:DIS) removing input synapse (synID="
	     << oldsyn->GetSynID() << ") from NeuronID=" << _neuronID << endl
	     << " (N:DIS) synapse has memory address: " << oldsyn << endl;
        #endif
	_inSynVect.erase(it); // erase it from _inSynVect!

	// now find synapse in incoming neutrons _outSynVect:
	syn_vect_t *vectptr = &oldsyn->GetFrom()->GetOutSynVect();
	for (it = vectptr->begin(); it != vectptr->end(); it++) {

	  if((*it)==oldsyn) {  // we found it.
            #ifdef NNET_DEBUG
	    cerr << " (N:DIS) removing output synapse (synID="
		 << oldsyn->GetSynID() << ") from NeuronID="
		 << oldsyn->GetFrom()->GetNeuronID() << endl;
            #endif

	    // erase it from _outSynVect!
	    vectptr->erase(it);
	  } // if((*it)==oldsyn) ...
	} // for(it) ...

	// finally free allocated memory
	delete oldsyn;
        #ifdef NNET_DEBUG
	cerr << " (N:DIS) allocated memory freed!" << endl;
        #endif

	return TRUE;
      } // if((*it)==oldsyn) ...
    } // for(it)...

    return FALSE; // syn is not an input synapse of this neutron!
    
  } // Neuron::DelInSyn(Synapse *)...


  // *** Neuron::DelInSyn(unsigned int) ***
  // searches for the synapse with synID in _inSynVect and removes it.
  // returns true if the synapse has been found and erased, false otherwise.
  inline bool Neuron::DelInSyn(unsigned int synID) {
    return DelInSyn(GetInSyn(synID));   // using other instance of DelInSyn()
  } // Neuron:DelInSyn(unsigned int) ...


  // *** Neuron::DelOutSyn(Synapse *) ***
  // searches for the synapse oldsyn in _inSynVect and removes it.
  // returns true if the synapse has been found and erased, false otherwise.
  bool Neuron::DelOutSyn(Synapse *oldsyn) {

    // check if it's an output synapse:
    for(syn_vect_t::iterator it = _outSynVect.begin();
	                    it != _outSynVect.end(); it++) {

      if((*it)==oldsyn) { // we found it.

        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (Neuron::DelOutSyn = N:DOS):" << endl
	     << " (N:DOS) removing output synapse (synID="
	     << oldsyn->GetSynID() << ") from NeuronID=" << _neuronID << endl
	     << " (N:DOS) synapse has memory address: " << oldsyn << endl;
        #endif
	_outSynVect.erase(it); // erase it from _outSynVect!

	// now find synapse in outgoing neutrons _inSynVect:
	syn_vect_t *vectptr = &oldsyn->GetTo()->GetInSynVect();
	for (it = vectptr->begin(); it != vectptr->end(); it++) {

	  if((*it)==oldsyn) {  // we found it.
            #ifdef NNET_DEBUG
	    cerr << " (N:DOS) removing input synapse (synID="
		 << oldsyn->GetSynID() << ") from NeuronID="
		 << oldsyn->GetTo()->GetNeuronID() << endl;
            #endif

	    // erase it from _inSynVect!
	    vectptr->erase(it);
	  } // if((*it)==oldsyn) ...
	} // for(it) ...

	// finally free allocated memory
	delete oldsyn;
        #ifdef NNET_DEBUG
	cerr << " (N:DOS) allocated memory freed!" << endl;
        #endif

	return TRUE;
      } // if((*it)==oldsyn) ...
    } // for(it)...

    return FALSE; // syn is not an output synapse of this neutron!
    
  } // Neuron::DelOutSyn(Synapse *)...


  // *** Neuron::DelOutSyn(unsigned int) ***
  // searches for the synapse with synID in _outSynVect and removes it.
  // returns true if the synapse has been found and erased, false otherwise.
  inline bool Neuron::DelOutSyn(unsigned int synID) {
      return DelOutSyn(GetOutSyn(synID)); // using other instance of DelInSyn()
  } // Neuron:DelOutSyn(unsigned int) ...


  // *** Neuron::AddInNeuron(Neuron&, unsigned int, double=default) ***
  // adds a neuron as input to _inSynVect creating the necessary synapse
  void Neuron::AddInNeuron(Neuron &newIn, unsigned int synID, double weight) {
    // default for weight is MAXDOUBLE, but we use DEF_INIT_WEIGHT as default:
    if (weight == MAXDOUBLE)
      weight = DEF_INIT_WEIGHT;

    Synapse *newSyn = new Synapse(synID, weight, &newIn, this);
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (Neuron::AddInNeuron = N:AIN):" << endl
	 << " (N:AIN) creating new Synapse(synID=" << synID 
	 << ", fromAdd=" << &newIn << ", toAdd=" << this << ")" << endl
	 << " (N:AIN) new synapse on memory address " << newSyn << endl;
    #endif
    newIn.GetOutSynVect().push_back(newSyn);
    #ifdef NNET_DEBUG
    cerr << " (N:AIN) connected as output synapse to NeuronID="
	 << newSyn->GetFrom()->GetNeuronID() 
	 << " (" << newSyn->GetFrom() << ")" << endl
	 << " (N:AIN) -> address added to _outSynVect is "
	 << newSyn->GetFrom()->GetOutSynVect().back() << endl;
    #endif
    _inSynVect.push_back(newSyn);
    #ifdef NNET_DEBUG
    cerr << " (N:AIN) connected as input synapse to NeuronID="
	 << newSyn->GetTo()->GetNeuronID() << " [this] ("
	 << newSyn->GetTo() << ")" << endl
	 << " (N:AIN) -> address added to _inSynVect is "
	 << newSyn->GetTo()->GetInSynVect().back() << endl;
    #endif
    
  } // Neuron::AddInNeuron(Neuron &, unsigned int, double=default) ...


  // *** Neuron::AddOutNeuron(Neuron &, unsigned int, double=defaulr) ***
  // adds a neuron as output to _outSynVect creating the necessary synapse
  void Neuron::AddOutNeuron(Neuron &newOut, unsigned int synID, double weight){
    // default for weight is MAXDOUBLE, but we use DEF_INIT_WEIGHT as default:
    if (weight == MAXDOUBLE)
      weight = DEF_INIT_WEIGHT;

    Synapse *newSyn = new Synapse(synID, weight, this, &newOut);
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (Neuron::AddOutNeuron = N:AON):" << endl
	 << " (N:AON) creating new Synapse(synID=" << synID 
	 << ", fromAdd=" << this << ", toAdd=" << &newOut << ")" << endl
	 << " (N:AON) new synapse on memory address " << newSyn << endl;
    #endif
    _outSynVect.push_back(newSyn);
    #ifdef NNET_DEBUG
    cerr << " (N:AON) connected as output synapse to NeuronID="
	 << newSyn->GetFrom()->GetNeuronID() << " [this] ("
	 << newSyn->GetFrom() << ")" << endl
	 << " (N:AON) -> address added to _outSynVect is "
	 << newSyn->GetFrom()->GetOutSynVect().back() << endl;
    #endif
    newOut.GetInSynVect().push_back(newSyn);
    #ifdef NNET_DEBUG
    cerr << " (N:AON) connected as input synapse to NeuronID=" << flush
	 << newSyn->GetTo()->GetNeuronID() << flush
	 << " (" << newSyn->GetTo() << ")" << endl << flush
	 << " (N:AON) -> address added to _inSynVect is " << flush
	 << newSyn->GetTo()->GetInSynVect().back() << endl;
    #endif
  } // Neuron::AddOutNeuron(Neuron &, unsigned int) ...


  //********************************************
  //*** class InputTerminal member functions ***
  //********************************************


  // *** InputTerminal::SetInput(double) ***
  // Accessor (non standard) for setting a new input value
  void InputTerminal::SetInput(double in) {

    if (_input!=in) {  // only if new input is different

      _input = in;
      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (InputTerminal::SetInput =IT:SI):" << endl
	   << " (IT:SI) NeuronID=" << _neuronID
	   << " (InputTerminal) new input: " << _input << endl;
      #endif

      // now inform the receiving neurons about the changed output
      // thus making sure that they will calculate their new output.
      for(syn_vect_t::iterator it=_outSynVect.begin();
	  it != _outSynVect.end(); it++) {

	((*it)->GetTo())->SetInputChange(TRUE);
        #ifdef NNET_DEBUG
	cerr << " (IT:SI) setting input flag of neuron with NeuronID=" 
	     << ((*it)->GetTo())->GetNeuronID() << " using synID="
	     << (*it)->GetSynID() << flush << endl;
        #endif
      } // for(vector ...

    } // if (_input!=in) ...

  } // InputTerinal::SetInput(double)


  // *** InputTerminal::AxonOutput(void) ***
  // redefiniton of virtual function AxonOutput, that in case of an input
  // terminal should only pass its input value to the output
  inline double InputTerminal::AxonOutput() {
    return _input;
  } // InputTerminal::AxonOutput(void)


  // ***********************************
  // *** class NNet member functions ***
  // ***********************************


  // *** Constructor: NNet::NNet(unsigned int, interm_map_t, neuron_map_t,
  //                             neuron_map_t, synapse_map_t) ***
  NNet::NNet(unsigned int nID, interm_map_t itmap, neuron_map_t nmap,
	     neuron_map_t outmap, synapse_map_t synmap) :
    _NNetID(nID), _ITMap(itmap), _NMap(nmap),
    _OutMap(outmap), _SynMap(synmap) {

    // determine maximum neuronID!
    _currNID = 0;
    for(interm_map_t::iterator it = itmap.begin(); it != itmap.end(); it++){
      if (_currNID < (*it).first)
	_currNID = (*it).first;
    }
    for(neuron_map_t::iterator it = nmap.begin(); it != nmap.end(); it++){
      if (_currNID < (*it).first)
	_currNID = (*it).first;
    }

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::NNet = NN:NN):" << endl
	 << " (NN:NN) current neuronID is " << _currNID << endl;
    #endif

    // determine maximum synID!
    _currSID = 0;
    for(synapse_map_t::iterator it = synmap.begin();
	it != synmap.end(); it++){
      if (_currSID < (*it).first)
	_currSID = (*it).first;
    }

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::NNet = NN:NN):" << endl
	 << " (NN:NN) current synID is " << _currSID << endl;
    #endif

  } // Constructor: NNet::NNet(unsigned int, interm_map_t, ...


  // *** Destructor: NNet::~NNet() *** (virtual)
  // The NNet class has virtual function members, thus this destructor is
  // virtual as well and may be redefined by derived classes.
  NNet::~NNet() {

    // first free all memory allocated for synapses:
    for(synapse_map_t::iterator it = _SynMap.begin();
	it != _SynMap.end(); it++) {
      delete (*it).second; 
    }

    // free all memory allocated for computing neurons:
    for(neuron_map_t::iterator it = _NMap.begin();
	it != _NMap.end(); it++) {
      delete (*it).second; 
    }

    // finally free all memory allocated for input terminals:
    for(interm_map_t::iterator it = _ITMap.begin();
	it != _ITMap.end(); it++) {
      delete (*it).second; 
    }

    // output terminals are part of the computing neurons, and therefore have
    // already been deleted!

  } // Destructor NNet::~NNet()


  // *** NNet::InsertInTerm(InputTerminal &, unsigned int) ***
  // This function inserts an InputTerminal to the network and returns
  // true if successful.
  // The function is virtual because it might be necessary to redefine
  // it for specific network structures.
  // If the neuron ID (nID) is not specified, the neuron ID will depend
  // on _currNID (default). 
  bool NNet::InsertInTerm(InputTerminal *inT, unsigned int nID) {
    // default value for nID is UINT_MAX!
    if (nID==UINT_MAX)
      nID = ++_currNID; // take neuron ID counter

    pair<unsigned int, InputTerminal*> interm;
    interm.first = nID;
    interm.second = inT;

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::InsertInTerm = NN:IIT):" << endl
	 << " (NN:IIT) inserting input terminal with nID="
	 << nID << endl;
    #endif

    if( (_ITMap.insert(interm)).second ){ // true if insert, false if not.
      if (nID > _currNID) _currNID = nID; // IMPORTANT: don't use IDs below nID
      return TRUE;
    }

    #ifdef NNET_DEBUG
    cerr << " (NN:IIT) input terminal couldn't be insert" << endl;
    #endif

    return FALSE; // could not insert
  } // NNet::InsertInTerm(InputTerminal &, unsigned int)


  // *** NNet::CreateInTerm(unsigned int=default, double=default) ***
  // CreateInTerm() creates a new InputTerminal for the network and
  // returns true if successful.
  // If the neuron ID (nID) is not specified, the neuron ID will depend
  // on _currNID (default). 
  bool NNet::CreateInTerm(unsigned int nID, double input) {
    // default value for nID is UINT_MAX!
    // default value for input is 0!
    if (nID==UINT_MAX)
      nID = ++_currNID; // take neuron ID counter

    InputTerminal *newIT = new InputTerminal(nID, input); // create terminal

    if(newIT)
      return InsertInTerm(newIT, nID);
    else
      return FALSE; // not enough memory
  } // NNet::CreateInTerm(unsigned int=default, double=default) ...


  // *** NNet::InsertNeuron(InputTerminal &, unsigned int) ***
  // This function inserts a computing neuron to the network and returns
  // true if successful.
  // The function is virtual because it might be necessary to redefine
  // it for specific network structures.
  // If the neuron ID (nID) is not specified, the neuron ID will depend
  // on _currNID (default). 
  bool NNet::InsertNeuron(Neuron *N, unsigned int nID) {
    // default value for nID is UINT_MAX!
    if (nID==UINT_MAX)
      nID = ++_currNID; // take neuron ID counter

    pair<unsigned int, Neuron*> neuron;
    neuron.first = nID;
    neuron.second = N;

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::InsertNeuron = NN:IN):" << endl
	 << " (NN:IN) inserting computing neuron with nID="
	 << nID << endl;
    #endif

    if( (_NMap.insert(neuron)).second ){ // true if insert, false if not.
      if (nID > _currNID) _currNID = nID; // IMPORTANT: don't use IDs below nID
      return TRUE;
    }

    #ifdef NNET_DEBUG
    cerr << " (NN:IN) computing neuron couldn't be insert" << endl;
    #endif

    return FALSE; // could not insert
  } // NNet::InsertNeuron(InputTerminal &, unsigned int)


  // *** NNet::CreateNeuron(unsigned int=default, double=default,
  //                        double=default, bool=default, double=default) ***
  // CreateNeuron() creates a new computing neuron for the network and
  // returns true if successful.
  // If the neuron ID (nID) is not specified, the neuron ID will depend
  // on _currNID (default). 
  bool NNet::CreateNeuron(unsigned int nID, double inSum, double outVal,
			  bool inCh, double delta) {
    // default value for nID is UINT_MAX!
    // default value for inSum is 0!
    // default value for outVal is 0!
    // default value for inCh is TRUE!
    // default value for delta is 0!

    if (nID==UINT_MAX)
      nID = ++_currNID; // take neuron ID counter

    // create computing neuron
    Neuron *newN = new Neuron(nID, inSum, outVal, inCh, delta);

    if(newN)
      return InsertNeuron(newN, nID);
    else
      return FALSE; // not enough memory
  } // NNet::CreateNeuron(unsigned int=default, double=default, double=default,
    //                    bool=default, double=default) ...


  // *** NNet::MakeOutput(unsigned int) ***
  // This function declares a computing neuron that is already part of the
  // network to be an output terminal and returns true if successful.
  // The function is virtual because it might be necessary to redefine
  // it for specific network structures.
  bool NNet::MakeOutput(unsigned int nID) {

    // first verify that nID is a computing neuron of this network
    neuron_map_t::iterator n_it = _NMap.find(nID);

    if( n_it != _NMap.end() ) {
      if( (_OutMap.insert(*n_it)).second ){ // true if not yet output terminal
        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (NNet::MakeOutput = NN:MO):" << endl
	     << " (NN:MO) declaring neuron with nID="
	     << nID << " as output" << endl;
        #endif
	return TRUE;
      }
      else { // already output
        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (NNet::MakeOutput = NN:MO):" << endl
	     << " (NN:MO) neuron with nID=" << nID
	     << " is already an output terminal" << endl;
        #endif
	return FALSE;
      }
    }

    // not a computing neuron => cannot be output
    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::MakeOutput = NN:MO):" << endl
	 << " (NN:MO) nID=" << nID << " is not a computing neuron, "
	 << "cannot make output" << endl;
    #endif

    return FALSE; // not a computing neuron
  } // NNet::MakeOutput(unsigned int nID)


  // *** NNet::CreateOutput(unsigned int=default) ***
  // CreateOutput() creates a new output terminal for the network and
  // returns true if successful. Important: Output terminals are
  // computing neurons as well!
  // If the neuron ID (nID) is not specified, the neuron ID will depend
  // on _currNID (default). 
  bool NNet::CreateOutput(unsigned int nID) {
    // default value for nID is UINT_MAX!
    if (nID==UINT_MAX)
      nID = ++_currNID; // take neuron ID counter

    Neuron *newO = new Neuron(nID); // create output terminal

    if( newO && InsertNeuron(newO, nID) ) 
      return MakeOutput(nID);
    else
      return FALSE; // not enough memory or not inserted in _OutMap
  } // NNet::CreateOutput(unsigned int=default) ...


  // *** NNet::Connect(unsigned int, unsigned int,
  //                   unsigned int=default, double=default) *** (virtual)
  // This function creates a synapse between two neurons:
  // The function is virtual because it might be necessary to redefine
  // it for specific network structures. If successful true is returned.
  bool NNet::Connect(unsigned int fromNID, unsigned int toNID,
		     unsigned int synID, double weight) {
    // The default value for synID is UINT_MAX
    // The default value for weight is MAXDOUBLE, but we have to use
    // DEF_INIT_WEIGHT instead:
    if (weight == MAXDOUBLE)
      weight = DEF_INIT_WEIGHT;

    interm_map_t::iterator from_it_it = _ITMap.find(fromNID);
    neuron_map_t::iterator from_n_it = _NMap.find(fromNID);
    neuron_map_t::iterator to_n_it = _NMap.find(toNID);

    // determine synID of new synapse
    unsigned int newSynID;
    if(synID==UINT_MAX)
      newSynID = ++_currSID; // default
    else {
      newSynID = synID;
      if (synID>_currSID) _currSID=synID;// IMPORTANT:don't use IDs below synID
    }


    if(to_n_it != _NMap.end()) { // found key toNID in _NMap

      if(from_it_it != _ITMap.end()) { // found key fromNID in _ITMap
                                       // first neuron is input terminal
        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (NNet::Connect = NN:C):" << endl
	     << " (NN:C) connecting nID=" << fromNID
	     << " (input terminal) to nID=" << toNID << endl
	     << " (NN:C) with synID=" << newSynID
	     << ", weight=" << weight << endl;
        #endif
	// connect neurons
	(*from_it_it).second->AddOutNeuron(*(*to_n_it).second,newSynID,weight);
	// add created synapse to _SynMap
	pair<unsigned int, Synapse*> syn;
	syn.first = newSynID;
	syn.second = (*from_it_it).second->GetOutSynVect().back();
	_SynMap.insert(syn);

	return TRUE;
      } // if(from_it_it != _NMap.end()) ...

      else if(from_n_it != _NMap.end()) { // found key toNID in _NMap

        #ifdef NNET_DEBUG
	cerr << "NNET_DEBUG (NNet::Connect = NN:C):" << endl
	     << " (NN:C) connecting nID=" << fromNID << " to nID=" << toNID
	     << endl << " (NN:C) with synID=" << newSynID
	     << ", weight=" << weight << endl;
        #endif
	// connect neurons
	(*from_n_it).second->AddOutNeuron(*(*to_n_it).second,newSynID,weight);
	// add created synapse to _SynMap
	pair<unsigned int, Synapse*> syn;
	syn.first = newSynID;
	syn.second = (*from_n_it).second->GetOutSynVect().back();
	_SynMap.insert(syn);

	return TRUE;
      } // else if(from_n_it != _NMap.end() ...

    } // if(to_n_it != _NMap.end()) ...

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::Connect = NN:C):" << endl
	 << " (NN:C) could not connect neurons with neuronIDs "
	 << fromNID << " and " << toNID << endl;
    #endif

    return FALSE; // did not connect the neurons
  } // NNet::Connect(unsigned int, unsigned int, unsigned int=default,
    //               double=default)


  // *** NNet::NNetInput(vector<double> &) ***
  // NNetInput writes a vector of input values to the input terminal,
  // returns true if successful.
  bool NNet::NNetInput(vector<double> &input) {

    // check if the size of the input vector corresponds to the number of
    // input terminals, return false otherwise.
    if( input.size() != _ITMap.size() ) return FALSE;

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::NNetInput = NN:NI):" << endl
	 << " (NN:NI) begin setting input values of input terminals" << endl;
    #endif

    // set input values of input terminals
    train_ex_in_t::iterator in_it = input.begin();
    for(interm_map_t::iterator it = _ITMap.begin();
	it != _ITMap.end(); it++, in_it++) {
      (*it).second->SetInput(*in_it);
    }

    #ifdef NNET_DEBUG
    cerr << " (NN:NI) input values of all input terminals set." << endl;
    #endif

    return TRUE;
  } // NNet::NNetInput(vector<double> &)


  // *** NNet::NNetOutput(vector<double> &) ***
  // NNetOutput() reads the nnet output to a vector of output values.
  bool NNet::NNetOutput(vector<double> &output) {
    
    // first clear output vector
    output.clear();

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::NNetOutput = NN:NO):" << endl
	 << " (NN:NO) begin reading output values of output terminals" << endl;
    #endif

    // copy output values of output terminals
    for(neuron_map_t::iterator it = _OutMap.begin();
	it != _OutMap.end(); it++) {
      output.push_back((*it).second->AxonOutput());
    }

    #ifdef NNET_DEBUG
    cerr << " (NN:NI) output values of all output terminals read." << endl;
    #endif

    return TRUE;
  } // NNet::NNetOutput(vector<double> &)


  // *** NNet::NNetIterate(void) *** (virtual)
  // NNetIterate() implements the iteration to generate the nnet output.
  // This function is virtual because it might depend by the structure
  // of the network, thus it might be necessary to overload it.
  void NNet::NNetIterate() {

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (NNet::NNetIterate = NN:It):" << endl
	 << " (NN:It) begin iteration of the neural network" << endl
	 << " (NN:It) calculate output values of computing neurons:" << endl;
    #endif
    
    // first compute the output of  all computing neurons (assuming that the
    // input terminals have already been set with NNetInput() - this does
    // already include the output neurons.
    for(neuron_map_t::iterator it = _NMap.begin(); it != _NMap.end(); it++) {
      (*it).second->AxonOutput();
    }

    #ifdef NNET_DEBUG
    cerr << " (NN:It) network iteration completed." << endl;
    #endif

  } // NNet::NNetIterate(void)


  // *************************************
  // *** class FFNNet member functions ***
  // *************************************


  // *** Destructor: FFNNet::~FFNNet() *** (virtual)
  // The FFNNet class has virtual function members, thus this destructor is
  // virtual as well and may be redefined by derived classes.
  FFNNet::~FFNNet() {

    // free all memory allocated for the layer vectors:
    for(vector<n_vect_t*>::iterator l_it = _nLayers.begin();
	l_it != _nLayers.end(); l_it++) {
      delete (*l_it);
    }

  } // destructor FFNNet::~FFNNet()


  // *** FFNNet::BuildNNet(void) ***
  // BuildNNet() returns true if the creation was successful.
  bool FFNNet::BuildNNet() {

    // make sure we won't do this twice!
    if(_netBuilt) {
      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (FFNNet::BuildNNet = FF:BN):" << endl
	   << " (FF:BN) cannot build network, network already built!"
	   << endl;
      #endif
      return FALSE;
    }

    // check if the network structure was already set (SetLayerWidthVect()) 
    if(_lWVec.size() < 2) { // we need at least an input and a computing layer!
      #ifdef NNET_DEBUG
      cerr << "NNET_DEBUG (FFNNet::BuildNNet = FF:BN):" << endl
	   << " (FF:BN) cannot build network, at least two layers needed."
	   << endl;
      #endif
      return FALSE;
    }

    // first allocate memory for all neurons, adding them to the corresponding
    // lists and maps:
    vector<unsigned int>::iterator width_it = _lWVec.begin();
    _currNID = 0;

    // allocate memory for input terminals (first layer is input!)
    for(unsigned int count = 0; count < (*width_it); count++) {
      // create and add to map of all input terminals
      if(!CreateInTerm(++_currNID)) return FALSE;
    }

    // allocate memory for computing neurons (including output terminals):
    for(width_it++; width_it != _lWVec.end(); width_it++) {

      // first allocate memory for layer vector:
      n_vect_t *newLayer = new n_vect_t;

      // allocate memory for computing neurons of this layer
      Neuron *newN;
      for(unsigned int count = 0; count < (*width_it); count++) {
	// create and add to map of all computing neurons, verify
	if( !(newN = new Neuron(++_currNID)) || !InsertNeuron(newN,_currNID) )
	  return FALSE;

	(*newLayer).push_back(newN); // add to layer vector
      }

      _nLayers.push_back(newLayer); // add layer vector to structure vector
      
    } // for(width_it++; ...

    // we do NOT allocate memory for output terminals, since the are already
    // part of the computing neurons


    // now connect input terminals to first computing layer (creating synapses)
    vector<n_vect_t*>::iterator nLayer_it = _nLayers.begin();

    for(interm_map_t::iterator it_it = _ITMap.begin(); // input terminal
	it_it != _ITMap.end(); it_it++) {

      for(n_vect_t::iterator neuron_it = (*nLayer_it)->begin(); // comp.neur.
	  neuron_it != (*nLayer_it)->end(); neuron_it++) {
	// connect input terminal (it_it) to computing neuron (neuron_it)
	// and add created synapse to map of all synapses
	if( !Connect((*it_it).first, (*neuron_it)->GetNeuronID()) ) return FALSE;
      } // for(n_vect_t ...

    } // for(interm_map_t ...

    // now connect all computing layers
    for(; (*nLayer_it) != _nLayers.back(); nLayer_it++) { 

      for(n_vect_t::iterator n_it1 = (*nLayer_it)->begin(); // current layer
	  n_it1 != (*nLayer_it)->end(); n_it1++) {

	nLayer_it++; // iterate next layer in the following for loop (n_it2)

	for(n_vect_t::iterator n_it2 = (*nLayer_it)->begin(); // next layer
	    n_it2 != (*nLayer_it)->end(); n_it2++) {
	  // connect neuron of current layer (n_it1)
	  // to neuron of next layer (n_it2)
	  // and add created synapse to map of all synapses, verify.
	  if( !Connect((*n_it1)->GetNeuronID(), (*n_it2)->GetNeuronID()) )
	    return FALSE;
	} // for(n_it2) ...

	nLayer_it--; // iterate current layer in this for loop (n_it1)

      } // for(n_it1) ...

    } // for(; (*nLayer_it) ...

    // finally let point _OutMap to the output terminals (last computing layer)
    for(n_vect_t::iterator n_it = (*nLayer_it)->begin(); 
	n_it != (*nLayer_it)->end(); n_it++) {
      // add to map of all output terminals
      MakeOutput((*n_it)->GetNeuronID());
    } // for(n_vect_t ...

    // network successfully built!
    _netBuilt = TRUE;
    return TRUE;
  } // FFNNet::BuildNNet(void)


  // *** FFNNet::BuildNNet(vector<unsigned int> &) ***
  // this overloaded BuildNNet function takes the network structure as
  // an argument, copies it to _lWVec and then calls BuildNNet(void) instance
  inline bool FFNNet::BuildNNet(vector<unsigned int> &lWV) {
    _lWVec = lWV;       // copy structure
    return BuildNNet();
  } // FFNNet::BuildNNet(vector<unsigned int> &) ...


  // *** FFNNet::NNetIterate(void) ***
  // NNetIterate() implements the iteration to generate the nnet output.
  void FFNNet::NNetIterate() {
    // this assumes that the input values of the input terminals have already
    // been set (NNetInput function member of class NNet)

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (FFNNet::NNetIterate = FF:It):" << endl
	 << " (FF:It) begin calculation of output values of computing neurons:"
	 << endl;
    #endif

    // calculate output values of all computing neurons (incl. output term.):
    for(vector<n_vect_t*>::iterator layer_it = _nLayers.begin();
	layer_it != _nLayers.end(); layer_it++) {          // neuron layers

      for(n_vect_t::iterator neuron_it = (*layer_it)->begin();
	  neuron_it != (*layer_it)->end(); neuron_it++) {     // single neurons

	(*neuron_it)->AxonOutput(); // compute output
      } // for(n_vect_t ...
    } // for(vector<n_vect_t*> ...

    #ifdef NNET_DEBUG
    cerr << " (FF:It) network iteration completed."
	 << endl;
    #endif

  } // FFNNet::NNetIterate(void)


  // ***************************************
  // *** class FFNNetBP member functions ***
  // ***************************************


  // *** FFNNetBP::BackPropagate(const train_ex_t &, double, double, int) ***
  // Implements the back-propagation algorithm (Bryson and Ho, 1969) for
  // supervisioned training of the neural network.
  // The training well be considered complete if the sum of the squared
  // differences (computed output values - expected output values) is lower
  // than the threshold (return value is treu) or if a maximum number of
  // iterations is reached (return value is false).
  // False will be returned also if an error occured.
  bool FFNNetBP::BackPropagate(const train_ex_t &tex, double learnrate,
			       double threshold, int maxIter) {

    train_ex_in_t  t_invec;   // input values of training example
    train_ex_out_t t_outvec;  // desired output values of training example
    train_ex_out_t netoutvec; // computed output values of neural network
    train_ex_out_t errvec;   // output error vector (netoutvec - t_outvec)

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (FFNNetBP::BackPropagate = FB:BP):" << endl;
    #endif

    // first check if the number of input and output values of the training
    // example fits the number of the input and output terminals of the
    // neural network:
    if( ( (tex.first->size())!=(GetITMap().size()) ) ||
	( (tex.second->size())!=(GetOutMap().size() ) ) ) {
      #ifdef NNET_DEBUG
      cerr << " (FB:BP) training example doesn't fit to no. of input and"
	   << " output terminals!" << endl;
      #endif
      return FALSE;
    } // if ...

    // get the input vector from the training example:
    #ifdef NNET_DEBUG
    cerr << " (FB:BP) input vector of training example:" << endl
	 << " (FB:BP) ";
    #endif
    for(train_ex_in_t::iterator it = tex.first->begin();
	it != tex.first->end(); it++) {
      #ifdef NNET_DEBUG
      cerr << (*it) << ", ";
      #endif
      t_invec.push_back(*it);
    }


    // get the vector of expected outputs from the training example:
    #ifdef NNET_DEBUG
    cerr << endl << " (FB:BP) output vector of training example:" << endl
	 << " (FB:BP) ";
    #endif
    for(train_ex_out_t::iterator it = tex.second->begin();
	it != tex.second->end(); it++) {
      #ifdef NNET_DEBUG
      cerr << (*it) << ", ";
      #endif
      t_outvec.push_back(*it);
    }

    #ifdef NNET_DEBUG
    cerr << endl << " (FB:BP) learnate = " << learnrate << ", threshold = "
	 << threshold << ", maxIter = " << maxIter << endl;
    #endif

    // set the input values to the input terminals:
    if(!NNetInput(t_invec)) {
      #ifdef NNET_DEBUG
      cerr << " (FB:BP) input terminals could not be set according to"
	   << " the training example" << endl;
      #endif
      return FALSE;
    }


    // **** main loop *** of back-propagation algorithm:

    double sq_error_sum; // sum of all squared output errors

    for(int iter = 1; iter <= maxIter; iter++) {

      #ifdef NNET_DEBUG
      cerr << endl << " (FB:BP) Iteration no. " << iter << endl;
      #endif

      // compute output of the neural network:
      NNetIterate();

      // get vector of computed output values (stored to netoutvec):
      if(!NNetOutput(netoutvec)) {
      #ifdef NNET_DEBUG
	cerr << " (FB:BP) could not determine the network output for the"
	     << " training example" << endl;
      #endif
	return FALSE;
      }


      // calculate the output error vector (t_outvec - netoutvec)
      // and sum all squared output errors:
      sq_error_sum = 0;
      double out_error;
      train_ex_out_t::iterator it = t_outvec.begin();

      errvec.clear(); // build a new one.

      for(train_ex_out_t::iterator n_it = netoutvec.begin();
	  n_it != netoutvec.end(); n_it++, it++){

	out_error = (*it)-(*n_it);
	errvec.push_back(out_error);

        #ifdef NNET_DEBUG
	cerr << " (FB:BP) training result: t_out - netout = " << (*it)
	     << " - " << (*n_it) << " = " << (*it)-(*n_it) << endl;
        #endif

	sq_error_sum += out_error * out_error;
      } // for(train...)


      #ifdef NNET_DEBUG
      cerr << " (FB:BP) sum of squared values of diff. vector: sq_error_sum = "
	   << sq_error_sum << endl;
      #endif

      // compare sq_error_sum to the threshold
      if(sq_error_sum<threshold) { // successful training, return TRUE
        #ifdef NNET_DEBUG
	cerr << " (FB:BP) training successful! " << iter-1
	     << " iteration";
	if((iter-1)!=1) cout <<"s";
	cout << " computed:" << endl
	     << " (FB:BP) sq_error_sum = " << sq_error_sum
	     << " < threshold = " << threshold << endl;
        #endif
	return TRUE;
      } // if(sq_error_sum<threshold)


      // determine new weights of synapses entering the different layers of
      // computing neurons, starting with the output layer:
      double delta, newWeight;

      #ifdef NNET_DEBUG
      cerr << " (FB:BP) determine new weights of all synapses:" << endl;
      #endif

      for(vector<n_vect_t*>::reverse_iterator layer_it = _nLayers.rbegin();
	  layer_it != _nLayers.rend(); layer_it++) { 

        #ifdef NNET_DEBUG
	cerr << " (FB:BP) new layer of computing neurons:" << endl;
        #endif

	if( layer_it == _nLayers.rbegin() ) // output terminal!
	  it = errvec.begin();              // we need the output error

	for(n_vect_t::iterator n_it = (*layer_it)->begin();
	    n_it != (*layer_it)->end(); n_it++) { // for each neuron

	  if( layer_it == _nLayers.rbegin() ) { // output terminal!
	    // delta = output error * gradient of input value
	    delta = (*n_it)->Delta(*it);
	    it++;
	  }
	  else { // internal neuron, don't pass an error value
	    // delta = gradient * sum of weighted deltas of output synapses
	    delta = (*n_it)->Delta();
	  }

          #ifdef NNET_DEBUG
	  if( layer_it == _nLayers.rbegin() ) // output terminal!
	    cerr << " (FB:BP) output terminal, ";
	  else                                // internal neuron
	    cerr << " (FB:BP) computing neuron, ";
	  cerr << "neuronID=" << (*n_it)->GetNeuronID()
	       << ", delta= " << delta <<endl;
          #endif

	  // adjust weight of input synapses of the current output neuron:
	  for(syn_vect_t::iterator syn_it = (*n_it)->GetInSynVect().begin();
	      syn_it != (*n_it)->GetInSynVect().end(); syn_it++) {
	    
	    newWeight = (*syn_it)->GetWeight() +
	              learnrate * delta * ((*syn_it)->GetFrom()->AxonOutput());
	    (*syn_it)->SetWeight(newWeight);
	  } // for(syn_vect_t ...
	} // for(n_vect_t ...

      } // for(<vector<n_vect_t*> ...


    } // for(int iter = 1; iter <= maxIter; iter++) ...
    // end of *** main loop ***


    // If we reach this statement the maximum number of iterations was computed

    #ifdef NNET_DEBUG
    cerr << " (FB:BP) training NOT successful! " << maxIter
	 << " iteration";
    if(maxIter!=1) cout <<"s";
    cout << " computed:" << endl
	 << " (FB:BP) sq_error_sum = " << sq_error_sum << " > threshold = "
	 << threshold << endl;
    #endif
    return FALSE;

  } // FFNNetBP::BackPropagate(const train_ex_t &, double, double, int) ...


  // *** FFNNetBP:: TrainNNet(train_set_t &, double, double, int) *** (virtual)
  // TrainNNet() implements a wrapper to the BackPropagation algorithm using
  // a whole set of training examples instead of only one.
  // This is a virtual function since it depends on the net structure.
  bool FFNNetBP::TrainNNet(train_set_t &tset, double learnrate,
			   double threshold, int maxIter) {

    for(train_set_t::iterator tset_it = tset.begin();
	tset_it != tset.end(); tset_it++) { // single training examples
      if(!BackPropagate((*tset_it), learnrate, threshold, maxIter))
	return FALSE; // BackPropagation() returned false, stop training
    }

    return TRUE; 
  } // FFNNetBP::TrainNNet(train_set_t &, double, double, int) ...


} // namespace nnet
