/*****************************************************************************
 * nnetfileio.cpp - library file for loading/saving neural networks (C++)    *
 * (libnnetfileio.so) or (libnnetfileio.a) - Version 0.1                     *
 * ------------------------------------------------------------------------- *
 * This library contains functions that load or save neural networks         *
 * based on the libnnet library.                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2003 by Rosario Piro (mailto: devfaust@users.sourceforge.net)   *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

// for debugging purpose:
#ifdef DEBUG
#define NNET_DEBUG
#endif

//*********************
//*** include files ***
//*********************
#include <nnetfileio.hpp>
#include <ctime>
#include <string>
#include <fstream>
#ifdef NNET_DEBUG
  #include <iostream>
#endif
#include <cstring>

//***************************
//*** preprocessor macros ***
//***************************
#define NNET_FILEFORMAT "nnetfile_type1"
#define IBL_ 2048   // input buffer length
// types of neural networks:
#define TYPE_NNET     "NNet"
#define TYPE_FFNNET   "FFNNet"
#define TYPE_FFNNETBP "FFNNetBP"

namespace nnet {

  // *************************
  // *** Helping functions ***
  // *************************

  // *** SaveGenNNet(NNet &, const string, const string) ***
  // Saves information that is contained in NNet and all its derived classes.
  // Will be called by all overloaded instances of SaveNNet().
  bool SaveGenNNet(NNet &net, string filename, const string nntype) {

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (SaveNNet = SNN):" << endl
	 << " (SNN) Writing network of type " << nntype << endl
	 << " (SNN) to file " << "'"<< filename << "'." << endl;
    #endif

    // overwrite existing file
    ofstream outFile(filename.c_str());
    if(!outFile) { // couldn't open
      #ifdef NNET_DEBUG
      cerr << " (SNN) error: cannot open file '" << filename << "'"
	   << " for writing!" << endl;
      #endif
      return FALSE;
    } // if(!outFile) ...


    // file header:
    time_t currTime = time(NULL);
    outFile << "# " << NNETLIB << ";" << NNETVER    // library name and version
	    << ";"  << NNET_FILEFORMAT << ";" << endl       // data file format
	    << "# date: " << ctime(&currTime)                   // current date
	    << "# IMPORTANT: don't modify this file!" << endl
	    << ".." << endl; // IMPORTANT: ".." indicates the begining of the
                             // data section

    // data section: first the net ID
    outFile << "NNetID=" << net.GetNNetID() << ";" 
	    << "NNetType=" << nntype << ";" << endl;


    // save information about input terminals (neuron ID, input, synapse
    // IDs of outgoing synapses):
    unsigned int ITsize = net.GetITMap().size(); 
    outFile << "ITsize=" << ITsize << ";" << endl; // number of input terminals
    // write one line for each input terminal
    for(interm_map_t::iterator mi = net.GetITMap().begin();
	mi != net.GetITMap().end(); mi++) {
      outFile << "neuronID=" << (*mi).second->GetNeuronID() 
	      << ";input=" << (*mi).second->AxonOutput()
	      << ";outSynIDs=";
      // synapse IDs of outgoing synapses
      syn_vect_t *outSV = &(*mi).second->GetOutSynVect();
      for(syn_vect_t::iterator vi = outSV->begin(); vi != outSV->end(); vi++) {
	outFile << (*vi)->GetSynID() << ","; // delimiter between elements
      } // for(vi) ...
      outFile << ";" << endl;
    } // for(mi) ...


    // now we save information about all computing neurons including output
    // terminals (neuron ID, sum of input values, status of input - changed 
    // or not -, output value, synapse IDs of incoming and outgoing vectors):
    unsigned int Nsize = net.GetNMap().size(); 
    outFile << "Nsize=" << Nsize << ";" << endl; // number of computing neurons
    // write one line for each computing neuron
    for(neuron_map_t::iterator mi = net.GetNMap().begin();
	mi != net.GetNMap().end(); mi++) {
      outFile << "neuronID=" << (*mi).second->GetNeuronID() 
	      << ";sum=" << (*mi).second->GetInSum() << ";inCh="
	      << (*mi).second->GetInputChange() << ";outVal="
	      << (*mi).second->GetOutputVal() << ";delta="
	      << (*mi).second->GetDeltaVal() << ";inSynIDs=";
      // synapse IDs of incoming synapses
      syn_vect_t *inSV = &(*mi).second->GetInSynVect();
      for(syn_vect_t::iterator vi = inSV->begin();
	  vi != inSV->end(); vi++) {
	outFile << (*vi)->GetSynID() << ",";
      } // for(vi) ...
      outFile << ";";
      // synapse IDs of outgoing synapses
      outFile << "outSynIDs=";
      syn_vect_t *outSV = &(*mi).second->GetOutSynVect();
      for(syn_vect_t::iterator vi = outSV->begin();
	  vi != outSV->end(); vi++) {
	outFile << (*vi)->GetSynID() << ",";
      } // for(vi) ...
      outFile << ";" << endl;
    } // for(mi) ...


    // output terminals are part of _NMap, thus we just have to save their
    // neuron ID in order to distinguish them from internal neurons:
    unsigned int Outsize = net.GetOutMap().size(); 
    outFile << "Outsize=" << Outsize << ";" << endl // no. of output terminals
	    << "neuronIDs=";
    for(neuron_map_t::iterator mi = net.GetOutMap().begin();
	mi != net.GetOutMap().end(); mi++) {
      // neuron IDs of output terminals
      outFile << (*mi).second->GetNeuronID() << ",";
    } // for(mi) ...
      outFile << ";" << endl;


    // finally we'll have to save all information of the synapses (synapse ID,
    // weight, neuron ID of incoming and outgoing neuron):
    unsigned int Synsize = net.GetSynMap().size(); 
    outFile << "Synsize=" << Synsize << ";" << endl; // number of synapses
    // write one line for each synapse
    for(synapse_map_t::iterator mi = net.GetSynMap().begin();
	mi != net.GetSynMap().end(); mi++) { 
      outFile << "synID=" << (*mi).second->GetSynID() 
	      << ";wgt=" << (*mi).second->GetWeight()
	      << ";fromID=" 
	      << (*mi).second->GetFrom()->GetNeuronID()
	      << ";toID=" 
	      << (*mi).second->GetTo()->GetNeuronID() << ";" 
	      << endl;
    } // for(mi) ...


    if(nntype==TYPE_FFNNET || nntype==TYPE_FFNNETBP) {
                        // this is an object of type FFNNet or FFNNetBP!
                        // save information specific to objects of class FFNNet

      // cast in order to access members of FFNNet!
      FFNNet *ffnnet = (FFNNet *) &net;

      outFile << "nLayers=" << ffnnet->GetNeuronLayers().size()
	      << ";" << endl;

      for(vector<n_vect_t*>::iterator layer_it =
	              ffnnet->GetNeuronLayers().begin();    // computing layer
	  layer_it != ffnnet->GetNeuronLayers().end(); layer_it++) {

	outFile << "neuronIDs=";
	for(n_vect_t::iterator neuron_it = (*layer_it)->begin();
	    neuron_it != (*layer_it)->end(); neuron_it++) { // single neurons
	  outFile << (*neuron_it)->GetNeuronID() << ",";
	} // for(n_vect_t ...

	outFile << ";" << endl;
      } // for(vector<n_vect_t*> ...

    } // if(nntype==TYPE_FFNNET || nntype==TYPE_FFNNET) ...

    // end of file
    outFile << "# end: " << NNETLIB << ";" << NNETVER << ";" << NNET_FILEFORMAT
	    << ";" << endl; // library name, version and file format

    // check for output errors
    if (outFile.bad() || outFile.fail()) {
      #ifdef NNET_DEBUG
      cerr << " (SNN) error while writing '" << filename << "'" << endl;
      #endif
      outFile.close();
      return FALSE;
    } 

    outFile.close();
    return TRUE; // no errors    
  } // SaveGenNNet(NNet &, const string, const string) ...


  // *******************************************************
  // *** overloaded functions for saving neural networks ***
  // *******************************************************

  // *** SaveNNet(NNet &, const string) ***
  // SaveNNet() saves the current state of an oject of type NNet to a file
  // returns true if successfully saved.
  inline bool SaveNNet(NNet &net, const string filename) {
    return SaveGenNNet(net, filename, TYPE_NNET);
  } // SaveNNet(NNet &, const string) ...


  // *** SaveNNet(FFNNet &, const string) ***
  // SaveNNet() saves the current state of an oject of type FFNNet
  // to a file returns true if successfully saved.
  inline bool SaveNNet(FFNNet &net, const string filename) {
    return SaveGenNNet(net, filename, TYPE_FFNNET);
  } // SaveNNet(FFNNet &, const string) ...


  // *** SaveNNet(FFNNetBP &, const string) ***
  // SaveNNet() saves the current state of an oject of type FFNNetBP
  // to a file returns true if successfully saved.
  inline bool SaveNNet(FFNNetBP &net, const string filename) {
    return SaveGenNNet(net, filename, TYPE_FFNNETBP);
  } // SaveNNet(FFNNetBP &, const string) ...


  // ********************************************
  // *** function for reading neural networks ***
  // ********************************************


  // *** NNet::LoadNNet(NNet &, const string) ***
  // LoadNNet() loads a nnet from a file, builds it and returns a pointer
  // to it, if successful. This pointer is a pointer to a GERNERIC NNet object
  // and needs to be typecasted by the programmer in order to access additional
  // information (e.g. of FFNNet or FFNetBP networks)!
  NNet *LoadNNet(const string filename) {

    NNet *net_ptr;
    
    unsigned int netID;
    string netType;

    // for parsing the input file
    char inBuffer[IBL_];
    unsigned int synID = 0;
    unsigned int neuronID = 0;

    #ifdef NNET_DEBUG
    cerr << "NNET_DEBUG (LoadNNet = LNN):" << endl;
    #endif

    // open nnet file
    ifstream inFile(filename.c_str());
    if(!inFile) { // couldn't open
      #ifdef NNET_DEBUG
      cerr << " (LNN) error: cannot open file '" << filename << "'"
	   << " for reading." << endl;
      #endif
      return NULL;
    } // if(!inFile) ...


    // first check file header (right file format)
    string nnetlib = NNETLIB;
    string searchstr = "# " + nnetlib;

    inFile.get(inBuffer,IBL_,';'); inFile.get(); // extract library name
    if ( inBuffer != searchstr ) {    // not a nnet file!
      #ifdef NNET_DEBUG
      cerr << " (LNN) error while reading '" << filename << "'" << endl
	   << " (LNN) not a nnet file!" << endl;
      #endif
      inFile.close();
      return NULL;
    }

    inFile.get(inBuffer,IBL_,';'); inFile.get(); // discard library version no


    // check file format version (NNET_FILEFORMAT)
    string nnetfileformat = NNET_FILEFORMAT;

    inFile.get(inBuffer,IBL_,';'); inFile.get();
    if( inBuffer != nnetfileformat ) {
      #ifdef NNET_DEBUG
      cerr << " (LNN) error while reading '" << filename << "'" << endl
	   << " (LNN) wrong format version! (should be: "
	   << NNET_FILEFORMAT << ")" << endl;
      #endif
      inFile.close();
      return NULL;
    }

    inFile.get(); // new line character
    inFile.get(inBuffer,IBL_); inFile.get(); // discard line containing date
    inFile.get(inBuffer,IBL_); inFile.get(); // discard third header line


    // search for delimiter between header and data section
    inFile.get(inBuffer,IBL_); inFile.get();
    if( strcmp(inBuffer,"..") ) {
      #ifdef NNET_DEBUG
      cerr << " (LNN) error while reading '" << filename << "'" << endl
	   << " (LNN) does not contain a data section!" << endl;
      #endif
      inFile.close();
      return NULL;
    }


    // now we assume that we can read the rest of the file, we will check
    // only afterwards!


    // get ID of neural network
    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "NNetID="
    inFile.get(inBuffer,IBL_,';'); inFile.get();
    netID = atoi(inBuffer);
    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "NNetType="
    inFile.get(inBuffer,IBL_,';'); inFile.get();
    inFile.get(); // new line
    netType = inBuffer;

    #ifdef NNET_DEBUG
    cerr << " (LNN) loading network from file '" << filename << "'." << endl
	 << " (LNN) netID=" << netID << ", netType=" << netType << endl;
    #endif


    // The memory we have to allocate depends on netType!!!
    if(netType==TYPE_FFNNETBP)
      net_ptr = (NNet*) new FFNNetBP(netID); // feed-forward with back prop.
    else if(netType==TYPE_FFNNET)
      net_ptr = (NNet*) new FFNNet(netID);   // feed-forward
    else
      net_ptr = new NNet(netID);             // default: generic network 

    if(!net_ptr) {
      #ifdef NNET_DEBUG
      cerr << " (LNN) error: not enough memory for neural network!" << endl;
      #endif
      inFile.close();
      return NULL;
    }

    // get information about input terminals (neuron ID, input, synapse
    // IDs of outgoing synapses):
    unsigned int ITsize;
    double inval;

    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "ITsize="
    inFile.get(inBuffer,IBL_); inFile.get();  // number of input terminals
    ITsize = atoi(inBuffer);

    #ifdef NNET_DEBUG
    cerr << " (LNN) reading input terminals, ITsize=" << ITsize << endl;
    #endif

    // read data lines of input terminals
    for(unsigned int count = 0; count < ITsize; count++) { // read each IT

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "neuronID="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); // read neuron ID 
      neuronID = atoi(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "input="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); // read input value
      inval = atof(inBuffer);

      #ifdef NNET_DEBUG
      cerr << " (LNN) neuronID=" << neuronID << ", input=" << inval << endl;
      #endif

      // create the input terminal and add it to the network
      if( !(net_ptr->CreateInTerm(neuronID,inval)) ) {
        #ifdef NNET_DEBUG
	cerr << " (LNN) error: could not create input terminal with neuronID="
	     << neuronID << endl;
        #endif
	inFile.close();
	return NULL;
      }

      // skip "outSynIDs=". Needed data for synapses will be read later.
      inFile.get(inBuffer,IBL_); inFile.get();

    } // for(count) ...


    // now we read information about all computing neurons (neuron ID, sum of
    // input values, status of input - changed or not -, output value, synapse
    // IDs of incoming and outgoing vectors):
    unsigned int Nsize;
    bool inch;
    double sum, outval, delta;

    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "Nsize="
    inFile.get(inBuffer,IBL_); inFile.get();  // number of computing neurons
    Nsize = atoi(inBuffer);

    #ifdef NNET_DEBUG
    cerr << " (LNN) reading computing neurons, Nsize=" << Nsize << endl;
    #endif

    // read one line for each computing neuron
    for(unsigned int count = 0; count < Nsize; count++) {

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "neuronID="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); 
      neuronID = atoi(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "sum="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); 
      sum = atof(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "inCh="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); 
      if(inBuffer=="1")
	inch = TRUE;
      else
	inch = FALSE;
      
      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "outVal="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); 
      outval = atof(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "delta="
      inFile.get(inBuffer,IBL_,';'); inFile.get(); 
      delta = atof(inBuffer);

      #ifdef NNET_DEBUG
      cerr << " (LNN) neuronID=" << neuronID << ", sum=" << sum
	   << ", inCh=" << inch << ", outVal=" << outval 
	   << ", delta=" << delta << endl;
      #endif

      // create the computing neuron and add it to the network
      if( !(net_ptr->CreateNeuron(neuronID,sum,outval,inch,delta)) ) {
        #ifdef NNET_DEBUG
	cerr << " (LNN) error: could not create computing neuron with "
	     << "neuronID=" << neuronID << endl;
        #endif
	inFile.close();
	return NULL;
      }

      // skip "inSynIDs=" and "outSynIDs="
      inFile.get(inBuffer,IBL_); inFile.get();

    } // for(count) ...
    

    // the output terminals are already included in the list of all computing
    // neurons, thus we just have to read their neuron ID inorder to
    // access them:
    unsigned int Outsize;

    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "Outsize="
    inFile.get(inBuffer,IBL_); inFile.get();  // number of output neurons
    Outsize = atoi(inBuffer);

    #ifdef NNET_DEBUG
    cerr << " (LNN) reading output terminals, Outsize=" << Outsize << endl;
    #endif

    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "neuronIDs="
    // neuron IDs of output terminals
    char nextchar;
    inFile.get(nextchar);
    while(nextchar!=';') { // another neuron ID to read
      inFile.putback(nextchar);
      inFile.get(inBuffer,IBL_,','); inFile.get();
      neuronID = atoi(inBuffer);

      #ifdef NNET_DEBUG
      cerr << " (LNN) output terminal with neuronID=" << neuronID << endl;
      #endif

      // declare the computing neuron with neuronID to be an output terminal
      if( !(net_ptr->MakeOutput(neuronID)) ) {
        #ifdef NNET_DEBUG
	cerr << " (LNN) error: could not declare neuron with neuronID="
	     << neuronID << " to be an output." << endl;
        #endif
	inFile.close();
	return NULL;
      }

      inFile.get(nextchar);
    } // while(nextchar!=';') ...

    inFile.get();


    // finally we'll have to read all information of the synapses (synapse ID,
    // weight, neuron ID of incoming and outgoing neuron):
    unsigned int Synsize;
    unsigned int from_nID, to_nID;
    double syn_w;

    inFile.get(inBuffer,IBL_,'='); inFile.get(); // "Synsize="
    inFile.get(inBuffer,IBL_); inFile.get();  // number of synapses
    Synsize = atoi(inBuffer);

    #ifdef NNET_DEBUG
    cerr << " (LNN) reading synapses, Synsize=" << Synsize << endl;
    #endif

    // read one line for each synapse
    for(unsigned int count = 0; count < Synsize; count++) { 
      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "synID="
      inFile.get(inBuffer,IBL_,';'); inFile.get();
      synID = atoi(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "wgt="
      inFile.get(inBuffer,IBL_,';'); inFile.get();
      syn_w = atof(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "fromID="
      inFile.get(inBuffer,IBL_,';'); inFile.get();
      from_nID = atoi(inBuffer);

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "toID="
      inFile.get(inBuffer,IBL_); inFile.get();
      to_nID = atoi(inBuffer);

      #ifdef NNET_DEBUG
      cerr << " (LNN) synID=" << synID << ", wgt=" << syn_w
	   << ", fromID=" << from_nID << ", toID=" << to_nID << endl;
      #endif

      // connect the corresponding neurons!
      if( !(net_ptr->Connect(from_nID,to_nID,synID,syn_w)) ) {
        #ifdef NNET_DEBUG
	cerr << " (LNN) error: could not create synapse (synID="
	     << synID << ") to connect" << endl << " (LNN) neuronID="
	     << from_nID << " to neuronID=" << to_nID << endl;
        #endif
	inFile.close();
	return NULL;
      }

    } // for(count) ...

    inFile.get(); // new line


    // we have read all information for generic NNet objects
    // now check which type of neural network this is:

    if( netType==TYPE_FFNNET || netType==TYPE_FFNNETBP) {
                                        // feed-forward, read additional info
      unsigned int noLayers;

      inFile.get(inBuffer,IBL_,'='); inFile.get(); // "nLayers="
      inFile.get(inBuffer,IBL_); inFile.get();  // number of computing layers
      noLayers = atoi(inBuffer);

      #ifdef NNET_DEBUG
      cerr << " (LNN) feed-forward network, reading " << noLayers
	   << " computing layers" << endl;
      #endif

      for(unsigned int count = 0; count < noLayers; count++) {// comp. layer

	n_vect_t *currentLayer = new n_vect_t;

        #ifdef NNET_DEBUG
        cerr << " (LNN) reading neuron IDs of new layer:" << endl
	     << " (LNN) neuronIDs=";
	#endif

	inFile.get(inBuffer,IBL_,'='); inFile.get(); // "neuronIDs="

	// neuron IDs of current layer
	char nextchar;
	neuron_map_t::iterator niter;

	inFile.get(nextchar);
	while(nextchar!=';') { // another neuron ID to read
	  inFile.putback(nextchar);
	  inFile.get(inBuffer,IBL_,','); inFile.get();
	  neuronID = atoi(inBuffer);

	  niter = net_ptr->GetNMap().find(neuronID);
	  if(niter == net_ptr->GetNMap().end()) { // not found!
            #ifdef NNET_DEBUG
	    cerr << endl << " (LNN) error: could not create layered "
		 << "structure." << endl << " (LNN) no computing neuron "
		 << "with neuronID=" << neuronID << endl;
            #endif
	    inFile.close();
	    return NULL;
	  }
	  currentLayer->push_back((*niter).second); // add pointer to layer

	  inFile.get(nextchar);

          #ifdef NNET_DEBUG
          cerr << neuronID << ",";
          #endif
	} // while(nextchar!=',') ...

	inFile.get();

        #ifdef NNET_DEBUG
	cerr << ";" << endl;
        #endif

	// add layer of vector pointers to the network
	// We have to typecast in order to access the corresponding members!!!

	if(netType==TYPE_FFNNETBP)  // feed-forward with back propagation
	  ((FFNNetBP*)net_ptr)->GetNeuronLayers().push_back(currentLayer);
	else                        // feed-forward
	  ((FFNNet*)net_ptr)->GetNeuronLayers().push_back(currentLayer);

      } // for(unsigned int ...

      inFile.get(); // new line

    } // if(netType==TYPE_FFNNET || netType==TYPE_FFNNETBP) ...

    // check final line of file
    inFile.get(inBuffer,IBL_,':'); inFile.get(); // "# end:"
    if (!strcmp(inBuffer, "# end")) {
      #ifdef NNET_DEBUG
      cerr << " (LNN) error: file not properly terminated." << endl;
      #endif
      inFile.close();
      return NULL;
    }


    // check for input errors
    if (inFile.bad() || inFile.fail()) {
      #ifdef NNET_DEBUG
      cerr << " (LNN) error while reading '" << filename << "'" << endl;
      #endif
      inFile.close();
      return NULL;
    }


    inFile.close(); // loading done.


    // FAUST: free all memory that isn't needed for the neural network

    return net_ptr;    

  } // NNet::LoadNNet(const string) ...



} // namespace nnet
