/*****************************************************************************
 * test_file.cpp - test program for neural network file I/O library (C++)    *
 * (libnnetfileio.so) or (libnnetfileio.a) - Version 0.1                     *
 * ------------------------------------------------------------------------- *
 * This program tests the library containing the basic classes and methods   *
 * for the implementation of simple neural networks. It tests the            *
 * the functionality of the network classes NNet, FFNNet and FFNNetBP.       *
 * ------------------------------------------------------------------------- *
 * Copyright 2003 by Rosario Piro (mailto: devfaust@users.sourceforge.net)   *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

#include <nnet.hpp>
#include <nnetfileio.hpp>
#include <iostream>

using namespace nnet;

int main() {

  bool failed = FALSE;

  cout << endl << "Allocating memory for an object of type FFNNetBP with netID"
       << " = 333." << endl;

  FFNNetBP *ffnnet = new FFNNetBP(333);
  if(!ffnnet) {
    failed = TRUE;
    return EXITBAD; // cannot go on ...
  }

  cout << "Building FFNNet with 4x4 neurons, the first layer being input"<<endl
       << "terminals and the last layer output terminals (-> three "<<endl
       << "layers of computing neurons)." << endl;

  vector<unsigned int> structure;
  for(int i=0; i <4; i++)
    structure.push_back(4);

  if( !ffnnet->BuildNNet(structure))
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Setting up new weights:" << endl;

  // setting weights of input synapses to computing neurons:
  double weights[]={0.8, 0.6, 0.2, 0.7,  // input syn. to neuron 1 of layer 1
		    1.0, 0.3, 1.0, 0.8,  // input syn. to neuron 2 of layer 1
		    0.2, 0.3, 1.0, 0.5,  // input syn. to neuron 3 of layer 1
		    0.4, 0.7, 0.7, 0.6,  // input syn. to neuron 4 of layer 1
		    0.3, 0.2, 0.4, 0.8,  // input syn. to neuron 1 of layer 2
		    0.8, 1.0, 0.2, 0.6,  // input syn. to neuron 2 of layer 2
		    0.9, 0.7, 0.6, 0.8,  // input syn. to neuron 3 of layer 2
		    0.2, 0.7, 0.8, 0.6,  // input syn. to neuron 4 of layer 2
		    0.4, 0.3, 0.6, 0.8,  // input syn. to neuron 1 of layer 3
		    0.2, 0.8, 0.7, 0.8,  // input syn. to neuron 2 of layer 3
		    0.8, 0.0, 1.0, 0.7,  // input syn. to neuron 3 of layer 3
		    1.0, 1.0, 1.0, 1.0}; // input syn. to neuron 4 of layer 3

  int w_count = 0;
  int lay_count = 1;
  for(vector<n_vect_t*>::iterator layer_it=ffnnet->GetNeuronLayers().begin();
      layer_it != ffnnet->GetNeuronLayers().end(); layer_it++){

    int n_count = 1;
    for(n_vect_t::iterator n_it = (*layer_it)->begin();
	n_it != (*layer_it)->end(); n_it++) {

      int syn_count = 1;
      for(syn_vect_t::iterator syn_it = (*n_it)->GetInSynVect().begin();
	  syn_it != (*n_it)->GetInSynVect().end(); syn_it++) {

	(*syn_it)->SetWeight(weights[w_count++]);
	cout << "Weight of input syn. " << syn_count << " of neuron "
	     << n_count << " in layer " << lay_count << " set to "
	     << (*syn_it)->GetWeight() << endl;
	syn_count++;
      }
      n_count++;
    }
    lay_count++;
  } // for(layer_it) ...


  cout << endl << "Setting input vector:  (1.0, 0.8, 1.0, 0.7)" << endl;
  train_ex_in_t inputvec;
  inputvec.push_back(1.0);
  inputvec.push_back(0.8);
  inputvec.push_back(1.0);
  inputvec.push_back(0.7);
  if ( !ffnnet->NNetInput(inputvec) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << endl << "Iterating network ..." << endl;
  ffnnet->NNetIterate();


  cout << endl
       << "Writing FFNNetBP to output file '_test_ffnnet_file.nnet'." << endl;
  if( !SaveNNet(*ffnnet, "_test_ffnnet_file.nnet") )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << "Writing the same object as a NNet to output file "
       << "'_test_net_file.nnet'." << endl
       << "File should not contain information about the layered structure"
       << endl;
  if( !SaveNNet( *((NNet*)ffnnet), "_test_nnet_file.nnet") )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << "Free the memory allocated for the object of type FFNNetBP."
       << endl;
  delete ffnnet;


  NNet *nn;
  FFNNetBP *ffn;

  cout << "Trying to read file '_test_nnet_file.nnet' (NNet)." << endl;
  if( !(nn=LoadNNet("_test_nnet_file.nnet")) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Writing the loaded network to output file '_test_nnet_file.nnet2'"
       << endl << "and running diff to compare the files:" << endl;
  if( !SaveNNet(*nn, "_test_nnet_file.nnet2") )
    failed = TRUE;
  else if( system("diff _test_nnet_file.nnet _test_nnet_file.nnet2") )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << "Trying to read file '_test_ffnnet_file.nnet' (FFNNet)." << endl;
  if( !(ffn=(FFNNetBP*)LoadNNet("_test_ffnnet_file.nnet")) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << "Writing the loaded network to output file '_test_ffnnet_file.nnet2'"
       << endl << "and running diff to compare the files:" << endl;
  if( !SaveNNet(*ffn, "_test_ffnnet_file.nnet2") )
    failed = TRUE;
  else if( system("diff _test_ffnnet_file.nnet _test_ffnnet_file.nnet2") )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  // conclusion
  if(failed) {
    cout << endl << "At least one of the tests did fail." << endl;
    return EXITBAD;
  }
 
  cout << endl << "Ok, Chummer! NNet file I/O test completed." << endl;
  return EXITGOOD;
}
