/*****************************************************************************
 * test_nnet.cpp - test program for basic neural network library (C++)       *
 * (libnnet.so) or (libnnet.a) - Version 0.1                                 *
 * ------------------------------------------------------------------------- *
 * This program tests the library containing the basic classes and methods   *
 * for the implementation of simple neural networks. It tests the            *
 * the functionality of the network classes NNet, FFNNet and FFNNetBP.       *
 * ------------------------------------------------------------------------- *
 * Copyright 2003 by Rosario Piro (mailto: devfaust@users.sourceforge.net)   *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

#include <nnet.hpp>
#include <iostream>

using namespace nnet;

int main() {

  interm_map_t ITmap; neuron_map_t Nmap; neuron_map_t Omap; synapse_map_t Smap;

  bool failed = FALSE;

  cout << "Testing the functionalities of class NNet:" << endl;
  cout << "Building a simple net of 5 input terminals and 10 computing neurons"
       << endl << "(5 of which are output terminals) ..." << endl;

  pair<unsigned int, InputTerminal*> interm;
  pair<unsigned int, Neuron*> neuron;

  for(int i = 0; i<10 ; i++) {
    if(i<5) {
      interm.first = i+1; interm.second = new InputTerminal(i+1);
      ITmap.insert(interm);
    }
    neuron.first = i+11; neuron.second = new Neuron(i+11);
    Nmap.insert(neuron);
  }
  // output terminals are part of the computing neurons
  for(neuron_map_t::iterator iter = Nmap.begin(); iter != Nmap.end();
      iter++, iter++) { // every second neuron
    Omap.insert(*iter);
  }

  cout << "No. of ITs = " << ITmap.size()
       << " (5), neurons = " << Nmap.size()
       << " (10), OTs = " << Omap.size()
       << " (5), synapses = " << Smap.size() << " (0)" << endl;
  if(ITmap.size()!=5 || Nmap.size()!=10 || Omap.size()!=5 || Smap.size()!=0)
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Now allocate memory for the NNet object and pass maps as" << endl
       << " arguments to the constructor. The synapse map is still empty!"
       << endl;

  NNet *nnet = new NNet(111, ITmap, Nmap, Omap, Smap);
  if(!nnet) {
    failed = TRUE;
    return EXITBAD; // cannot go on
  }

  // create 50 synapses IT->N
  cout << "Create 50 synapses IT->N:" << endl;
  for(interm_map_t::iterator it_it = nnet->GetITMap().begin();
       it_it != nnet->GetITMap().end(); it_it++) {
    for(neuron_map_t::iterator n_it = nnet->GetNMap().begin();
	n_it != nnet->GetNMap().end(); n_it++) {
      nnet->Connect((*it_it).first, (*n_it).first);
    }
  }

  cout << "no. of synapses is now " << nnet->GetSynMap().size()<<" (50)"<<endl;
  if(nnet->GetSynMap().size() != 50)
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Creating a new input terminal with neuronID=123." << endl;
  if( !nnet->CreateInTerm(123) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Creating a new computing neuron with neuronID=222." << endl;
  if( !nnet->CreateNeuron(222) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Creating a new output terminal (computing neuron!) with"
       << " neuronID=777." << endl;
  if ( !nnet->CreateOutput(777) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "_currNID is now: " << nnet->GetCurrentNeuronID() <<" (777)"<<endl
       <<endl; 
  if (nnet->GetCurrentNeuronID() != 777)
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Number of input terminals now: " << nnet->GetITMap().size()
       << " (6)" << endl << "Last one has neuronID="
       << (*nnet->GetITMap().rbegin()).second->GetNeuronID()<<" (123)" << endl
       <<endl;
  if( nnet->GetITMap().size() != 6 ||
      (*nnet->GetITMap().rbegin()).second->GetNeuronID() != 123 )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Number of output terminals now: " << nnet->GetOutMap().size()
       << " (6)" << endl << "Last one has neuronID="
       << (*nnet->GetOutMap().rbegin()).second->GetNeuronID()<<" (777)"<<endl
       << endl;
  if( nnet->GetOutMap().size() != 6 ||
      (*nnet->GetOutMap().rbegin()).second->GetNeuronID() != 777 )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Number of computing neurons now: " << nnet->GetNMap().size()
       << " (12)" << endl << "Last one has neuronID="
       << (*nnet->GetNMap().rbegin()).second->GetNeuronID()<<" (777)" <<endl
       << endl;
  if( nnet->GetNMap().size() != 12 ||
      (*nnet->GetNMap().rbegin()).second->GetNeuronID() != 777 )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Making computing neuron with neuronID=222 an output terminal."
       << endl;
  if( !nnet->MakeOutput(222) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Number of output terminals now: " << nnet->GetOutMap().size()
       << " (7)" << endl << "Last one has neuronID="
       << (*nnet->GetOutMap().rbegin()).second->GetNeuronID()<<" (777)"<<endl
       << endl;
  if( nnet->GetOutMap().size() != 7 ||
      (*nnet->GetOutMap().rbegin()).second->GetNeuronID() != 777 )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Making computing neuron with neuronID=222 an output terminal."
       << endl;
  if( nnet->MakeOutput(222) ) // IS already output -> should return false!
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Number of output terminals now: " << nnet->GetOutMap().size()
       << " (7)" << endl << "Last one has neuronID="
       << (*nnet->GetOutMap().rbegin()).second->GetNeuronID() <<" (777)"<<endl;
  if ( nnet->GetOutMap().size() != 7 ||
       (*nnet->GetOutMap().rbegin()).second->GetNeuronID() != 777 )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Making computing neuron with neuronID=999 an output terminal."
       << endl;
  if( nnet->MakeOutput(999) ) // 999 doesn't exist -> we expect false!
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Iterating network:" << endl;
  nnet->NNetIterate();
  cout << "iteration completed." << endl;
  cout << "Now reading output (should be zero):" << endl;
  vector<double> output_vect;
  if( !nnet->NNetOutput(output_vect) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Output vector (should be seven times zero): ";
  double outsum = 0;
  for(vector<double>::iterator it = output_vect.begin();
      it != output_vect.end(); it++) {
    cout << (*it) << ", ";
    outsum += (*it);
  }
  cout << endl << endl;
  if(outsum!=0)
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Creating an input vector (1,2,3,4,5,6):" << endl;
  vector<double> input_vect;
  for(int i=1; i<7; i++)
    input_vect.push_back(i);
  cout << "Setting input to network." << endl;
  if( !nnet->NNetInput(input_vect) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Iterating network:" << endl;
  nnet->NNetIterate();
  cout << "iteration completed." << endl;
  cout << "Now reading output:" << endl;
  if ( !nnet->NNetOutput(output_vect) )
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Output vector (last two should be zero, terminals have no synaptic "
       << "links): " << endl;
  for(vector<double>::iterator it = output_vect.begin();
      it != output_vect.end(); it++)
    cout << (*it) << ", ";
  cout << endl << endl;
  if( (output_vect[5]+output_vect[6]) != 0)
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;


  cout << "Invoking destructor NNet::~NNet for the created network." << endl;
  delete nnet;


  cout << endl << "Now checking member funtions of class FFNNet/FFNNetBP:"
       << endl;

  cout << endl << "Allocating memory for an object of type FFNNetBP." << endl;

  FFNNetBP *ffnnet = new FFNNetBP(333);
  if(!ffnnet) {
    failed = TRUE;
    return EXITBAD; // cannot go on ...
  }

  cout << "Building FFNNet with 4x4 neurons, the first layer being input"<<endl
       << "terminals and the last layer output terminals (-> three "<<endl
       << "layers of computing neurons)." << endl;

  vector<unsigned int> structure;
  for(int i=0; i <4; i++)
    structure.push_back(4);

  if( !ffnnet->BuildNNet(structure))
    failed = TRUE;
  else
    cout << "test ok" << endl << endl;

  cout << "Setting up new weights:" << endl;

  // setting weights of input synapses to computing neurons:
  double weights[]={0.8, 0.6, 0.2, 0.7,  // input syn. to neuron 1 of layer 1
		    1.0, 0.3, 1.0, 0.8,  // input syn. to neuron 2 of layer 1
		    0.2, 0.3, 1.0, 0.5,  // input syn. to neuron 3 of layer 1
		    0.4, 0.7, 0.7, 0.6,  // input syn. to neuron 4 of layer 1
		    0.3, 0.2, 0.4, 0.8,  // input syn. to neuron 1 of layer 2
		    0.8, 1.0, 0.2, 0.6,  // input syn. to neuron 2 of layer 2
		    0.9, 0.7, 0.6, 0.8,  // input syn. to neuron 3 of layer 2
		    0.2, 0.7, 0.8, 0.6,  // input syn. to neuron 4 of layer 2
		    0.4, 0.3, 0.6, 0.8,  // input syn. to neuron 1 of layer 3
		    0.2, 0.8, 0.7, 0.8,  // input syn. to neuron 2 of layer 3
		    0.8, 0.0, 1.0, 0.7,  // input syn. to neuron 3 of layer 3
		    1.0, 1.0, 1.0, 1.0}; // input syn. to neuron 4 of layer 3

  for(int test_no = 0; test_no < 4; test_no++) { // we always start with the
                                                 // network.
    int w_count = 0;

    int lay_count = 1;
    for(vector<n_vect_t*>::iterator layer_it=ffnnet->GetNeuronLayers().begin();
	layer_it != ffnnet->GetNeuronLayers().end(); layer_it++){

      int n_count = 1;
      for(n_vect_t::iterator n_it = (*layer_it)->begin();
	  n_it != (*layer_it)->end(); n_it++) {

	int syn_count = 1;
	for(syn_vect_t::iterator syn_it = (*n_it)->GetInSynVect().begin();
	    syn_it != (*n_it)->GetInSynVect().end(); syn_it++) {

	  (*syn_it)->SetWeight(weights[w_count++]);
	  cout << "Weight of input syn. " << syn_count << " of neuron "
	       << n_count << " in layer " << lay_count << " set to "
	       << (*syn_it)->GetWeight() << endl;
	  syn_count++;
	}
	n_count++;
      }
      lay_count++;
    } // for(layer_it) ...


    cout << endl << endl << "Now testing back-propagation:"
	 << endl << "Input vector of training example:  (1.0, 0.8, 1.0, 0.7)"
	 << endl;
    train_ex_in_t inputvec;
    inputvec.push_back(1.0);
    inputvec.push_back(0.8);
    inputvec.push_back(1.0);
    inputvec.push_back(0.7);

    cout<< "Output vector of training example: (0.9, 0.5, 0.4, 0.8)" << endl;
    train_ex_out_t outputvec;
    outputvec.push_back(0.9);
    outputvec.push_back(0.5);
    outputvec.push_back(0.4);
    outputvec.push_back(0.8);

    train_ex_t trainex;
    trainex.first = &inputvec;
    trainex.second = &outputvec;

    if (test_no == 0) { // first test ... no back propagation, threshold ok
      cout << "Starting back propagation with learnrate=0.8, threshold=0.5, "
	   << "maxIter=1" << endl;
      if( !ffnnet->BackPropagate(trainex, 0.8, 0.5, 1) ) // should return true!
	failed = TRUE;
      else
	cout << "test ok" << endl << endl;
    }
    else if(test_no == 1) { // second test ... 1 iteration, threshold not ok.
      cout << "Starting back propagation with learnrate=0.8, threshold=0.01, "
	   << "maxIter=1" << endl;
      if( ffnnet->BackPropagate(trainex, 0.8, 0.01, 1) ) // should return false
	failed = TRUE;
      else
	cout << "test ok" << endl << endl;
    }
    else if(test_no == 2) { // third test ... 10 iterations, threshold ok.
      cout << "Starting back propagation with learnrate=0.8, threshold=0.05, "
	   << "maxIter=10" << endl;
      if( !ffnnet->BackPropagate(trainex, 0.8, 0.05, 10) )//should return true
	failed = TRUE;
      else {
	cout << "Number of necessary iterations should be 5." << endl;
	cout << "test ok" << endl << endl;
      }
    }
    
    else if(test_no == 3) { // fourth test ... training set!.
      cout << "Starting back propagation with a training set of two examples"
	   << endl;
      train_set_t tset;
      tset.push_back(trainex); // first example is same as above!
      // now generate a new example
      train_ex_in_t inputvec2;
      inputvec2.push_back(0.7);
      inputvec2.push_back(0.2);
      inputvec2.push_back(0.6);
      inputvec2.push_back(0.7);
      train_ex_out_t outputvec2;
      outputvec2.push_back(0.8);
      outputvec2.push_back(0.5);
      outputvec2.push_back(0.1);
      outputvec2.push_back(0.4);

      trainex.first = &inputvec2;
      trainex.second = &outputvec2;
      tset.push_back(trainex); // push back second example!


      if( !ffnnet->TrainNNet(tset, 0.8, 0.05, 10) ) //should return true
	failed = TRUE;
      else {
	cout << "The first call of BackPropagate() should return true after 6"
	     << endl <<"iterations, the second one after 9 iterations.";
	cout << endl << "test ok" << endl << endl;
      }
    }


  } // for(test_no) ...

  cout << "Finally free the memory allocated for the object of type FFNNetBP."
       << endl;
  delete ffnnet;


  // conclusion
  if(failed) {
    cout << endl << "At least one of the tests did fail." << endl;
    return EXITBAD;
  }
 
  cout << endl << "Ok, Chummer! NNet network test completed." << endl;
  return EXITGOOD;
}
